program Mainfunc2D

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!  2D massively parallel HSS multifontal Helmholtz solver: HelmPar2D  !!!!!!!!!!!!!!!
!!!!!!!!!!!           Shen Wang,  Jianlin Xia,  Maarten V. de Hoop              !!!!!!!!!!!!!!!
!!!!!!!!!!!                Purdue & ConocoPhillips, June 2010                   !!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

use NDpre2D
use context
use A2D 
use fact
use solution2D
implicit none
include 'mpif.h'


!! declaring variables
integer                      ::  ierr, myrank, nprocs
integer                      ::  JOB, tos
integer                      ::  Lvl, slvl, Lpar, rkmin, mr2d
real(kind=4)                 ::  tol
real(kind=8)                 ::  t1, t2
integer                      ::  Nz, Nx, N_delta, NRHS
complex(kind=4)              ::  freq, sigma0, hz, hx
integer,allocatable          ::  sourceZ(:), sourceX(:)
real(kind=4),allocatable     ::  vel_pz(:,:), vel_sz(:,:), rho(:,:), epslon(:,:), delta(:,:), theta(:,:)
complex(kind=4),allocatable  ::  X(:,:,:)
real(kind=4)                 ::  tmp
integer                      ::  rhs, zz, xx, pp
character(len=200)           ::  model_vel, model_rho, model_epslon, model_delta, model_theta, filename 


!! MPI Initialization
call MPI_Init(ierr)
call MPI_Comm_rank( MPI_COMM_WORLD, myrank, ierr )
call MPI_Comm_size( MPI_COMM_WORLD, nprocs, ierr )

!! choosing jobs. 1: isotropic acoustic, 2: elliptic anisotropy, 3: VTI, 4. HTI, 5. TTI
JOB = 1
if (JOB == 5) then
    tos = 2
else
    tos = 1
end if

!! set up global variables
Lvl            = 16
slvl           = 0
rkmin          = 20                                       ! minimum rank to guarantee the numerical stability
tol            = 1.0E-4
Lpar           = nint(log(real(nprocs))/log(2.0)) + 1     ! Lpar = log_2(nprocs) + 1
mr2d           = 0
NRHS           = 1
Nz             = 2048
Nx             = 2048
N_delta        = 25
sigma0         = 2.0*pi*(60.0E+0,0.0E+0)
hz             = (15.0E+0,0.0E+0)
hx             = (15.0E+0,0.0E+0)
freq           = 2.0*pi*(26.0E+0,0.0E+0)
model_vel      = "/autohome/u111/wang273/data/BP2007/Vp_Model.dat"
model_rho      = ""
model_epslon   = "/autohome/u111/wang273/data/BP2007/Epsilon_Model.dat"
model_delta    = ""
model_theta    = "/autohome/u111/wang273/data/BP2007/Theta_Model.dat"
filename       = "~/Hsolver/f90/test2D.dat"

allocate(sourceZ(NRHS),sourceX(NRHS))
sourceZ = 0
sourceX = 0
do rhs = 1,NRHS
    sourceZ(rhs) = Nz*2/3 + (rhs-1)*1
    sourceX(rhs) = Nx*2/3 + (rhs-1)*1
end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  1. 2D nested dissection preprocessing  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
t1 = MPI_Wtime()
call NDpre_generate( myrank, tos, Lvl, Lpar, Nz, Nx, lengbTT, lenlocTT, gbTT, locTT, gbTlabel, locTlabel, gbNB, locNB, locHZ, locTZ, locHX, locTX, label, flag )
t2 = MPI_Wtime()
if (myrank == 0) then
    write(*,*) "2D nested dissection domain decomposition is done! wall time is ", t2-t1
end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  2. BLACS contexts generating  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
call ctxt_generate( flag, lengbTT, gbTT, ctxt, nprow, npcol, myrow, mycol, mynodes ) 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

if (myrank==1) then
  write(*,*) "locHX locTX locHZ locTZ", locHX, locTX,locHZ,locTZ
end if

!! read in velocity model
allocate(vel_pz(locTZ-locHZ+1,locTX-locHX+1))
if (1 < 0) then
    open( unit = 10, file = model_vel, form = "unformatted", status = "old", access = "direct", recl = 1 )
    do xx = locHX,locTX
        do zz = locHZ,locTZ
            read( 10, rec = (xx-1)*Nz+zz ) vel_pz(zz-locHZ+1,xx-locHX+1)
        end do
    end do
    close(10)
else
    vel_pz = 4000.0E+0
end if

!! vel_sz
allocate(vel_sz(locTZ-locHZ+1,locTX-locHX+1))
vel_sz = 0.0E+0

!! read in density model: CAUTION, different from other model parameters
allocate(rho(locTZ-locHZ+1+2*tos,locTX-locHX+1+2*tos))
if (1 < 0) then
    rho = 1.0E+0
    open( unit = 10, file = model_rho, form = "unformatted", status = "old", access = "direct", recl = 1 )
    do xx = locHX-tos,locTX+tos
        do zz = locHZ-tos,locTZ+tos
            if (xx >= 1 .and. xx <= Nx .and. zz >= 1 .and. zz <= Nz) then
                read( 10, rec = (xx-1)*Nz+zz ) rho(zz-locHZ+1+tos,xx-locHX+1+tos)
            end if
        end do
    end do
    close(10)
else
    rho = 1.0E+0
end if

!! read in epsilon model
allocate(epslon(locTZ-locHZ+1,locTX-locHX+1))
if (1 < 0) then
    open( unit = 10, file = model_epslon, form = "unformatted", status = "old", access = "direct", recl = 1 )
    do xx = locHX,locTX
        do zz = locHZ,locTZ
            read( 10, rec = (xx-1)*Nz+zz ) epslon(zz-locHZ+1,xx-locHX+1)
        end do
    end do
    close(10)
else
    epslon = 0.25E+0
end if

!! read in delta model
allocate(delta(locTZ-locHZ+1,locTX-locHX+1))
if (1 < 0) then
    open( unit = 10, file = model_delta, form = "unformatted", status = "old", access = "direct", recl = 1 )
    do xx = locHX,locTX
        do zz = locHZ,locTZ
            read( 10, rec = (xx-1)*Nz+zz ) delta(zz-locHZ+1,xx-locHX+1)
        end do
    end do
    close(10)
else
    delta = 0.0E+0
end if

!! read in theta model
allocate(theta(locTZ-locHZ+1,locTX-locHX+1))
if (1 < 0) then
    open( unit = 10, file = model_theta, form = "unformatted", status = "old", access = "direct", recl = 1 )
    do xx = locHX,locTX
        do zz = locHZ,locTZ
            read( 10, rec = (xx-1)*Nz+zz ) theta(zz-locHZ+1,xx-locHX+1)
        end do
    end do
    close(10)
else
    theta = 40.0E+0
end if


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  3. generating the matrix A  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
t1 = MPI_Wtime()

if (JOB == 1) then        ! isotropic acoustic
    call A2D_acoustic( freq, tos, Nz, Nx, N_delta, sigma0, hz, hx, locHZ, locTZ, locHX, locTX, label, vel_pz, rho, A )
elseif (JOB == 2) then    ! elliptic anisotropy
    call A2D_elliptic( freq, tos, Nz, Nx, N_delta, sigma0, hz, hx, locHZ, locTZ, locHX, locTX, label, vel_pz, rho, epslon, theta, A )
elseif (JOB == 3) then    ! VTI
    call      A2D_VTI( freq, tos, Nz, Nx, N_delta, sigma0, hz, hx, locHZ, locTZ, locHX, locTX, label, vel_pz, rho, epslon, delta, A )
elseif (JOB == 4) then    ! HTI
    call      A2D_HTI( freq, tos, Nz, Nx, N_delta, sigma0, hz, hx, locHZ, locTZ, locHX, locTX, label, vel_pz, rho, epslon, delta, A )
elseif (JOB == 5) then
    call      A2D_TTI( freq, tos, Nz, Nx, N_delta, sigma0, hz, hx, locHZ, locTZ, locHX, locTX, label, vel_pz, vel_sz, rho, epslon, delta, theta, A )
end if

t2 = MPI_Wtime()

if (myrank == 0) then
    write(*,*) "local matrix A generation is done! Wall time is", t2-t1
end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

deallocate( vel_pz, vel_sz, rho, epslon, delta, theta )    ! no use at all for later on solver


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  4. factorization  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
t1 = MPI_Wtime()
call factorization(A,slvl,Lpar,rkmin,tol,mr2d,flag,ctxt,nprow,npcol,myrow,mycol,mynodes, & 
                   lenlocTT,lengbTT,locTT,gbTT,locTlabel,gbTlabel,locNB,gbNB,locUL,locUR,locLL,locIPIV,gbUL,gbUR,gbLL,gbIPIV)
t2 = MPI_Wtime()
if (myrank == 0) then
    write(*,*) "The total factorization is done! Wall time is", t2-t1
end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!! creating an empty file
if (myrank == 0) then
    open( unit = 10, file = filename, form = "unformatted", status = "replace", access = "direct", recl = 1 )
    close(10)
end if

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  5. solution  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
allocate(X(locTZ-locHZ+1,locTX-locHX+1,NRHS))
X = ZERO
do rhs = 1,NRHS
    do xx = locHX,locTX
        do zz = locHZ,locTZ
            tmp = real((zz-sourceZ(rhs))**2+(xx-sourceX(rhs))**2)
            if (tmp <= 25.0E+0) then
                X(zz-locHZ+1,xx-locHX+1,rhs) = exp(-tmp/8.0)
            end if
        end do
    end do
end do
if (locHZ > 1) then   ! CAUTION!!
    X(1:tos,:,:) = ZERO
end if
if (locHX > 1) then
    X(:,1:tos,:) = ZERO
end if

t1 = MPI_Wtime()
call mfsolve(X,NRHS,slvl,Lpar,flag,Nz,Nx,locHZ,locTZ,locHX,locTX,label,ctxt,nprow,npcol,myrow,mycol,mynodes, &
             lenlocTT,lengbTT,locTT,gbTT,locTlabel,gbTlabel,locNB,gbNB,locUL,locUR,locLL,locIPIV,gbUL,gbUR,gbLL,gbIPIV)
t2 = MPI_Wtime()
if (myrank == 0) then
    write(*,*) "The wall time of the solution to", NRHS, " RHS is ", t2-t1
end if

! write only the real part of the wavefield
do pp = 0,nprocs-1
    if (myrank == pp) then
        call swriteX( filename, real(X), NRHS, tos, Nz, Nx, locHZ, locTZ, locHX, locTX ) 
    end if
    call MPI_Barrier(MPI_COMM_WORLD,ierr)
end do
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  6. deallocating factors  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
call factors_dealloc( lenlocTT, locUL, locUR, locLL, locIPIV, Lpar-1, gbUL, gbUR, gbLL, gbIPIV )
deallocate( locUL, locUR, locLL, locIPIV, gbUL, gbUR, gbLL, gbIPIV )
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!! MPI Finalization
call MPI_Finalize(ierr)


end program Mainfunc2D
