program adjtest

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!  2D massively parallel HSS multifontal Helmholtz solver: HelmPar2D  !!!!!!!!!!!!!!!
!!!!!!!!!!!           Shen Wang,  Jianlin Xia,  Maarten V. de Hoop              !!!!!!!!!!!!!!!
!!!!!!!!!!!                Purdue & ConocoPhillips, June 2010                   !!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

use NDpre2D
use context
use A2D 
use fact
use solution2D
use parser
use fileio
use powernorm

implicit none
include 'mpif.h'


!! declaring variables
integer                      ::  ierr, myrank, nprocs
integer                      ::  JOB, tos, passive
integer                      ::  Lvl, slvl, Lpar, rkmin, mr2d
real(kind=4)                 ::  tol, dt, dw, Jsum
real(kind=8)                 ::  t1, t2
integer                      ::  Nz, Nx, N_delta, NRHS, Nt, Nw,Nh,lnx,lnz,Np,nsi,Ng, fps=20
 complex(kind=4)              :: sigma0, hz, hx, tmpi,tmpdi,tmpdx,tmpdz,tmpg,tmpgpx,tmpgmx,tmpgpz,tmpgmz,dPhi,tmpss
integer,allocatable          ::  sourceZ(:), sourceX(:),h(:),arrayloc(:)
real(kind=4),allocatable     ::  Istack(:),Is(:),I0(:),Iadj(:),Ipr(:,:,:),dIpr(:,:,:)
real(kind=4),allocatable     ::  vel(:),vel0(:), dv(:), ga(:), gz(:),gx(:), taper(:), N(:),water(:),waterpr(:),rayp(:)
real(kind=4),allocatable     ::  Jig(:),Jall(:)
 complex(kind=4),allocatable  ::  Y(:),Y0(:),YY(:), freq(:),Aw(:),dataw(:),fieldout(:),fieldout0(:)
 complex(kind=4),allocatable  ::  X(:,:,:),dX(:,:,:) 
real(kind=4)                 ::  tmp, tmpz, tmpx,alpha,maxp,xis(1000000),zis(1000000),si1(1000000),si2(1000000)
integer                      ::  rhs, zz, xx, pp, iw, w0, wf,xi,zi, ig,ds,dh,ihh,hh, iss, ix, iz, headZ, headX,forward,adjyn
 character(len=256)           ::  vfile,v0file,dvfile,sfile,sfile0,rfile,rfile0,ifile,ifile0,wnum,pnum 
 !!complex(kind=4),parameter   :: pi = (3.1415926E+0,0.0E+0)
 !!complex(kind=4),parameter   :: ONE = (1.0E+0,0.0E+0), ZERO = (0.0E+0,0.0E+0), RT = (0.0E+0,1.0E+0)
real(kind=4),allocatable     ::  vel_pz(:,:), dvel_pz(:,:), rho(:,:)

!! MPI Initialization
 call MPI_Init(ierr)
 call MPI_Comm_rank( MPI_COMM_WORLD, myrank, ierr )
 call MPI_Comm_size( MPI_COMM_WORLD, nprocs, ierr )

if (myrank==0) write(*,*) "nprocs is ",nprocs

!! set up variables
 call frompar('level',Lvl,14)
 call frompar('Nz',Nz,2048)
 call frompar('Nx',Nx,2048)
 !call frompar('Ns',NRHS,1)
 call frompar('Nt',Nt,2048)
 call frompar('Nw',Nw,55)
 call frompar('Nh',Nh,20)
 call frompar('dh',dh,2)
 call frompar('dx',tmp,250.); hx=tmp
 call frompar('dz',tmp,250.); hz=tmp
 call frompar('dt',dt,0.01)
 call frompar('ds',ds,10)
 call frompar('w0',w0,2)
 call frompar('wf',wf,Nw)
 call frompar('x0',xi,80)
 call frompar('z0',zi,110)
 call frompar('alpha',alpha,1.)
 call frompar('vfile',vfile)
 call frompar('v0file',v0file)
 call frompar('dvfile',dvfile)
 call frompar('sfile',sfile0)
 call frompar('rfile',rfile0,' ')
 call frompar('ifile',ifile0)
 call frompar('passive',passive,0)
 call frompar('slownesses',Np,15)
 call frompar('maxp',maxp,30.)
 call frompar('forward',forward,1)
 call frompar('adjoint',adjyn,1) 

!! choosing jobs. 1: isotropic acoustic, 2: elliptic anisotropy, 3: VTI, 4. HTI, 5. TTI
JOB = 1
tos = 1


ng=1




if (myrank==0) then
write(*,*) "Data source: ",trim(sfile0),myrank
write(*,*) "Image file:  ",trim(ifile0)
endif

if (passive == 1) then
NRHS = Np*2 - 1
else
NRHS=int((Nx-2*xi)/ds + 1)
endif
if (myrank==0) write(*,*) "Number of sources", NRHS


allocate(Y(Nz*Nx*NRHS*2))
allocate(Is(Nz*Nx*NRHS*2))
allocate(Iadj(Nz*Nx*NRHS))
allocate(dataw(Nx*NRHS))
allocate(freq(Nw),Aw(Nw))
allocate(sourceZ(NRHS),sourceX(NRHS),rayp(NRHS))
allocate(vel(Nz*Nx))
allocate(vel0(Nz*Nx))
allocate(dv(Nz*Nx))
allocate(taper(Nx),arrayloc(Nx))
allocate(ga(ng), gx(ng), gz(ng))
allocate(fieldout(Nz*Nx*4),fieldout0(Nz*Nx*4))
!allocate(h(2*Nh-1))


dw = 2*PI/(Nt*dt)
do iw=1,Nw
  freq(iw)=(iw-1)*dw
  Aw(iw) = exp(-(real(wf+w0)/2 - real(iw))**2/(real(wf-w0)/5)**2)
enddo
sigma0 = 1.0*freq(wf)

sourceZ = 0
sourceX = 0
do rhs = 1,NRHS
    sourceZ(rhs) = zi 
    sourceX(rhs) = xi + ds*(rhs-1)
end do

NRHS = NRHS * 2 !! NOTE: TO ACCOMODATE SOURCE AND RECEIVER

    !!Vary strength of smooth perturbation to test error fn
    !!ga=(/-1., -0.5, 0., 0.5, 1.5, 2.0, 2.5, 3./) 
    
    
!! read in velocity model
  !call realinb(dv,Nx*Nz,vfile)
  call realinb(vel0,Nx*Nz,v0file)
  call realinb(dv,Nx*Nz,dvfile)
  do xx=1,Nx*Nz
    !!vel(xx)=vel0(xx)+ga(ig)*dv(xx) !Error function test
    vel0(xx)=vel0(xx)+alpha*dv(xx)
    !vel0(xx)=vel0(xx) !+alpha*dv(xx)  !!Starting model
    dv(xx)=alpha*dv(xx)
    
    !vel0(xx)=vel0(xx) + alpha  !!Add constant
    
    
    
  enddo
  if (myrank==0) write(*,*) "Loaded velocity models"


arrayloc = 0
!do xx=100,Nx-100,20
do xx=100,Nx-100,1
  arrayloc(xx)=1
enddo

taper=0.
do zz=1,Nx
  if (arrayloc(zz)==1) then
    do xx=1,Nx
      taper(xx)=taper(xx)+exp(-real(xx-zz)**2/2**2.)
    enddo
  endif
enddo
taper = taper/maxval(taper,Nx)

 ifile = 'taperout'
 call realoutb(taper,Nx,ifile)
 deallocate(arrayloc)


Is=0.


!! set up global variables
!Lvl            = 16
slvl           = 0
rkmin          = 20                                       ! minimum rank to guarantee the numerical stability
tol            = 1.0E-4
Lpar           = nint(log(real(nprocs))/log(2.0)) + 1     ! Lpar = log_2(nprocs) + 1
mr2d           = 0
N_delta        = 50



 if (.false.) then
   ihh=1
   gx(ihh)=160
   gz(ihh)=150
   ga(ihh)=0.
   
   do ix=160,560,50
     do iz=150,350,50
       ihh=ihh+1
       gx(ihh)=ix
       gz(ihh)=iz
       !ga(ihh)=(-1)**ihh*100
       ga(ihh)=-dv(iz+Nz*(ix-1))/1.8/2
     enddo
   enddo
 endif     




   ! gx=(/210., 210., 310., 410., 510., 210., 310., 410., 510./)
   ! gz=(/160., 160., 160., 160., 160., 260., 260., 260., 260./)
    !ga=(/  0., 100., 100., 100., 100., 100., 100., 100., 100./) 
   ! ga=(/-200., 200., -200., 200., 200., -200., 200., -200./) /200
    !!ga=(/0., 0., 0., 0., 200., -200., 200., -200./) 
    !vel = 7200.0E+0
    !!dv = 0.0
    
    gx=360.
    gz=170.
    ga=alpha
    
    do ig=1,Ng
    do xx=1,Nx
    	do zz=1,Nz
	
	   tmp = real((zz-gz(ig))**2+(xx-gx(ig))**2)
	   !vel0(zz+Nz*(xx-1)) = vel0(zz+Nz*(xx-1)) + ga(ig)*exp(-tmp/50.**2)
	   !vel(zz+Nz*(xx-1)) = vel(zz+Nz*(xx-1)) + ga(ig)*exp(-tmp/50.**2)
	   !vel(zz+Nz*(xx-1)) = vel(zz+Nz*(xx-1)) + ga(ig)*exp(-tmp/33.**2)
	enddo
    enddo
    enddo



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  1. 2D nested dissection preprocessing  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
t1 = MPI_Wtime()
 call NDpre_generate( myrank, tos, Lvl, Lpar, Nz, Nx, lengbTT, lenlocTT, gbTT, locTT, gbTlabel, locTlabel, gbNB, locNB, locHZ, locTZ, locHX, locTX, label, flag )
t2 = MPI_Wtime()
if (myrank == 0) then
    write(*,*) "2D nested dissection domain decomposition is done! wall time is ", t2-t1
end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  2. BLACS contexts generating  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 call ctxt_generate( flag, lengbTT, gbTT, ctxt, nprow, npcol, myrow, mycol, mynodes ) 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!! find vel model for processor
  allocate(vel_pz(locTZ-locHZ+1,locTX-locHX+1))
    allocate(dvel_pz(locTZ-locHZ+1,locTX-locHX+1))
    do xx = locHX,locTX
        do zz = locHZ,locTZ
	    vel_pz(zz-locHZ+1,xx-locHX+1) = vel0(zz+Nz*(xx-1))  !starting model
	    !dvel_pz(zz-locHZ+1,xx-locHX+1) = dv(zz+Nz*(xx-1))  !vel(zz,xx)
	    dvel_pz(zz-locHZ+1,xx-locHX+1) = vel(zz+Nz*(xx-1))  !! Starting model plus smooth pert. 	    
        end do
    end do
  
  allocate(rho(locTZ-locHZ+1+2*tos,locTX-locHX+1+2*tos))
  rho = 1.0E+0


  allocate(Ipr(locTZ-locHZ+1,locTX-locHX+1,NRHS))
  Ipr=0.
  allocate(dIpr(locTZ-locHZ+1,locTX-locHX+1,NRHS))
  dIpr=0.





if (forward==1) then  !!! DONT SKIP FORWARD PROPAGATION


do iw=w0,wf

!! Define initial conditions

Y = ZERO
!!Y0 = ZERO

if (myrank==0) then
write(*,*) " "
write(*,*) "Propagating for frequency ",iw
write(*,*) "********************************"
endif



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  3. generating the matrix A  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
t1 = MPI_Wtime()
if (JOB == 1) then        ! isotropic acoustic
    call A2D_acoustic( freq(iw), tos, Nz, Nx, N_delta, sigma0, hz, hx, locHZ, locTZ, locHX, locTX, label, vel_pz, rho, A )
endif
!t2 = MPI_Wtime()

if (myrank == 0) then
    !write(*,*) "local matrix A generation is done! Wall time is", t2-t1
end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  4. factorization  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!t1 = MPI_Wtime()
 call factorization(A,slvl,Lpar,rkmin,tol,mr2d,flag,ctxt,nprow,npcol,myrow,mycol,mynodes, & 
                   lenlocTT,lengbTT,locTT,gbTT,locTlabel,gbTlabel,locNB,gbNB,locUL,locUR,locLL,locIPIV,gbUL,gbUR,gbLL,gbIPIV)
!t2 = MPI_Wtime()
if (myrank == 0) then
    !write(*,*) "The total factorization is done! Wall time is", t2-t1
end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  4A. define initial wavefields !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  write(wnum,fmt='(I2)')iw
  sfile=trim(sfile0)//trim(wnum)
  call complexinb(dataw,Nx*NRHS/2,sfile)

  do rhs=1,NRHS/2
    do xx=1,Nx
      do zz=2,Nz-1
         tmpz = real((zz-sourceZ(rhs))**2)
	 tmpx = real((xx-sourceX(rhs))**2)
	 if (tmpz<=16.0) then
	                                                                                
            Y(sourceZ(rhs)+1+Nz*(xx-1)+Nz*Nx*(rhs-1+NRHS/2))= Aw(iw)*taper(xx)*exp(-tmpz/16.0)*conjg(dataw((xx-1)+Nx*(rhs-1)))
            Y(sourceZ(rhs)-1+Nz*(xx-1)+Nz*Nx*(rhs-1+NRHS/2))=-Aw(iw)*taper(xx)*exp(-tmpz/16.0)*conjg(dataw((xx-1)+Nx*(rhs-1)))	    
            !Y(sourceZ(rhs)+Nz*(xx-1)+Nz*Nx*(rhs-1+NRHS/2))=Aw(iw)*taper(xx)*exp(-tmpz/16.0)*freq(iw)*conjg(dataw((xx-1)+Nx*(rhs-1)))
	 endif
	enddo
    enddo
  enddo 
    !if (myrank==0) write(*,*) "debug5",maxval(abs(Y),Nz*Nx*NRHS)
if (passive == 1) then
  write(wnum,fmt='(I2)')iw
  sfile=trim(sfile0)//'s'//trim(wnum)
  call complexinb(dataw,Nx*NRHS/2,sfile)
  do rhs=1,NRHS/2
    do xx=1,Nx
      do zz=2,Nz-1
         tmpz = real((zz-sourceZ(rhs))**2)
	 tmpx = real((xx-sourceX(rhs))**2)
         if (tmpz <= 25.0E+0) then
            !Y(zz+Nz*(xx-1)+Nz*Nx*(rhs-1)) = (dataw((xx-1)+Nx*(rhs-1)))*exp(-(tmpz)/4.0)!*Aw(iw)
	    Y(zz+Nz*(xx-1)+Nz*Nx*(rhs-1)) = (dataw((xx-1)+Nx*(rhs-1)))*exp(-(tmpz)/4.0)!*Aw(iw)
         end if
      enddo
    enddo
  enddo
	 
else 
   
  do rhs=1,NRHS/2
    do xx=1,Nx
      do zz=2,Nz-1
         tmpz = real((zz-sourceZ(rhs))**2)
	 tmpx = real((xx-sourceX(rhs))**2)
         if (tmpx+tmpz <= 25.0E+0) then
	    Y(zz+Nz*(xx-1)+Nz*Nx*(rhs-1)) = Aw(iw)*exp(-(tmpx+tmpz)/4.0)	    
         end if
	enddo
    enddo
  enddo
endif
	
  Y=Y*100 !AVOID STUPID PROBLEMS 
	 

  !if (myrank==0) write(*,*) "debug6",maxval(abs(Y),Nz*Nx*NRHS)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  5. solution  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
lnx=locTX-locHX+1
lnz=locTZ-locHZ+1
allocate( X(lnz,lnx,NRHS))
allocate(dX(lnz,lnx,NRHS))
allocate(YY(lnz*lnx*NRHS))


X = ZERO
dX = ZERO
do rhs = 1,NRHS
    do xx = locHX,locTX
        do zz = locHZ,locTZ
	    X(zz-locHz+1,xx-locHx+1,rhs) = Y(zz+Nz*(xx-1)+Nz*Nx*(rhs-1))  !!Y(zz,xx,rhs)
	    dX(zz-locHz+1,xx-locHx+1,rhs) = Y(zz+Nz*(xx-1)+Nz*Nx*(rhs-1))  !! Compute unperturbed field
	    
        end do
    end do
end do
if (locHZ > 1) then   ! CAUTION!!
    X(1:tos,:,:) = ZERO
end if
if (locHX > 1) then
    X(:,1:tos,:) = ZERO
end if

!t1 = MPI_Wtime()
 call mfsolve(X,NRHS,slvl,Lpar,flag,Nz,Nx,locHZ,locTZ,locHX,locTX,label,ctxt,nprow,npcol,myrow,mycol,mynodes, &
             lenlocTT,lengbTT,locTT,gbTT,locTlabel,gbTlabel,locNB,gbNB,locUL,locUR,locLL,locIPIV,gbUL,gbUR,gbLL,gbIPIV)
!t2 = MPI_Wtime()
if (myrank == 0) then
    !write(*,*) "The wall time of the solution to", NRHS, " RHS is ", t2-t1
end if



if (.false.) then
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  3B. generating the matrix A  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 call factors_dealloc( lenlocTT, locUL, locUR, locLL, locIPIV, Lpar-1, gbUL, gbUR, gbLL, gbIPIV )
 deallocate(locUL, locUR, locLL, locIPIV, gbUL, gbUR, gbLL, gbIPIV )  
 
 t1 = MPI_Wtime()
if (JOB == 1) then        ! isotropic acoustic
    call A2D_acoustic( freq(iw), tos, Nz, Nx, N_delta, sigma0, hz, hx, locHZ, locTZ, locHX, locTX, label, dvel_pz, rho, A )
endif
!t2 = MPI_Wtime()

if (myrank == 0) then
    !write(*,*) "local matrix A generation is done! Wall time is", t2-t1
end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  4B. factorization  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!t1 = MPI_Wtime()
 call factorization(A,slvl,Lpar,rkmin,tol,mr2d,flag,ctxt,nprow,npcol,myrow,mycol,mynodes, & 
                   lenlocTT,lengbTT,locTT,gbTT,locTlabel,gbTlabel,locNB,gbNB,locUL,locUR,locLL,locIPIV,gbUL,gbUR,gbLL,gbIPIV)
!t2 = MPI_Wtime()
if (myrank == 0) then
    !write(*,*) "The total factorization is done! Wall time is", t2-t1
end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  5B. Perturbed Solution  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

do rhs = 1,NRHS
    do xx=1,locTX-locHX+1
        do zz = 1,locTZ-locHZ+1
	    !!dX(zz,xx,rhs) = 2*freq(iw)**2*vel_pz(zz,xx)**(-3)*dvel_pz(zz,xx)* X(zz,xx,rhs)
	enddo
    end do
end do
if (locHZ > 1) then   ! CAUTION!!
    dX(1:tos,:,:) = ZERO
end if
if (locHX > 1) then
    dX(:,1:tos,:) = ZERO
end if

t1 = MPI_Wtime()
 call mfsolve(dX,NRHS,slvl,Lpar,flag,Nz,Nx,locHZ,locTZ,locHX,locTX,label,ctxt,nprow,npcol,myrow,mycol,mynodes, &
             lenlocTT,lengbTT,locTT,gbTT,locTlabel,gbTlabel,locNB,gbNB,locUL,locUR,locLL,locIPIV,gbUL,gbUR,gbLL,gbIPIV)
t2 = MPI_Wtime()
if (myrank == 0) then
    write(*,*) "The wall time of the unperturbed solution is ", t2-t1
end if

endif
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



 allocate(waterpr(NRHS/2),water(NRHS/2))
 waterpr = 0.
 water = 0.
 do xx=1,locTX-locHX+1
   do zz = 1,locTZ-locHZ+1
      do rhs=1,NRHS/2
        tmp = 0.2*real(X(zz,xx,rhs)*conjg(X(zz,xx,rhs)))
        if (waterpr(rhs) .lt. tmp) then
           waterpr(rhs)=tmp
        endif
      enddo
   enddo
 enddo
 
 call MPI_ALLREDUCE(waterpr,water,NRHS/2,MPI_REAL,MPI_MAX,MPI_COMM_WORLD,ierr)
 if (myrank==0) then
    ifile = trim(ifile0)//'water'
    open( unit=fps,file=TRIM(ifile), access='direct',recl=NRHS/2,form='unformatted')
    write(unit=fps,rec=(iw))water
    close(fps)
 endif
  !if (myrank == 0) write(*,*) 'debugw', water(1),water(29)
  
!t1 = MPI_Wtime()
  do xx=1,lnx
   do zz = 1,lnz
    do rhs=1,NRHS/2
      
      
      tmpg = (X(zz,xx,rhs)) / (conjg(X(zz,xx,rhs)) * X(zz,xx,rhs) + water(rhs))
      tmpgpx = (X(zz,xx+1,rhs)) / (conjg(X(zz,xx+1,rhs)) * X(zz,xx+1,rhs) + water(rhs))
      tmpgmx = (X(zz,xx-1,rhs)) / (conjg(X(zz,xx-1,rhs)) * X(zz,xx-1,rhs) + water(rhs))
      tmpgpz = (X(zz+1,xx,rhs)) / (conjg(X(zz+1,xx,rhs)) * X(zz+1,xx,rhs) + water(rhs))
      tmpgmz = (X(zz-1,xx,rhs)) / (conjg(X(zz-1,xx,rhs)) * X(zz-1,xx,rhs) + water(rhs))
      
      

      tmpi = (X(zz,xx,rhs+NRHS/2)) *  tmpg !!(X(zz,xx,rhs)) !tmpg
      tmpdi = (X(zz,xx,rhs+NRHS/2)) *  (X(zz,xx,rhs)) !tmpg
      !tmpdi = (conjg(X(zz,xx,rhs)) * X(zz,xx,rhs) + water(rhs))
	    
      if (xx.gt.1 .and. xx.lt.lnx) then
        tmpdx = (tmpgpx-tmpgmx)*(X(zz,xx+1,rhs+NRHS/2)-X(zz,xx-1,rhs+NRHS/2))/(4*hx**2) 
      else if (xx.eq.1) then
        tmpdx = (tmpgpx-tmpg)*(X(zz,xx+1,rhs+NRHS/2)-X(zz,xx,rhs+NRHS/2))/(hx**2) 
      else if (xx.eq.lnx)  then      
        tmpdx = (tmpg-tmpgmx)*(X(zz,xx,rhs+NRHS/2)-X(zz,xx-1,rhs+NRHS/2))/(hx**2) 
      endif
      if (zz.gt.1 .and. zz.lt.lnz) then
        tmpdz = (tmpgpz-tmpgmz)*(X(zz+1,xx,rhs+NRHS/2)-X(zz-1,xx,rhs+NRHS/2))/(4*hz**2)
      else if (zz.eq.1) then
        tmpdz = (tmpgpz-tmpg)*(X(zz+1,xx,rhs+NRHS/2)-X(zz,xx,rhs+NRHS/2))/(hz**2) 
      else if (zz.eq.lnz)  then      
        tmpdz = (tmpg-tmpgmz)*(X(zz,xx,rhs+NRHS/2)-X(zz-1,xx,rhs+NRHS/2))/(hz**2) 
      endif
      
      
     
      
      Ipr(zz,xx,rhs) = Ipr(zz,xx,rhs) + real((PI*2*RT*freq(iw))**(-1)*(tmpi - vel_pz(zz,xx)**2/freq(iw)**2*(tmpdx + tmpdz)))
      !!Ipr(zz,xx,rhs) = Ipr(zz,xx,rhs) + real((PI*2*RT*freq(iw))**(-1)*(tmpi))
      !dIpr(zz,xx,rhs) = dIpr(zz,xx,rhs) + real((PI*2*RT*freq(iw))**(-1)*(tmpdi))
      !Ipr(zz,xx,rhs) = Ipr(zz,xx,rhs) + real((PI*2*RT*freq(iw))**(-1)*(tmpi - vel_pz(zz,xx)**2/freq(iw)**2*(tmpdx + tmpdz)))
      !dIpr(zz,xx,rhs) = dIpr(zz,xx,rhs) + real((PI*2*RT*freq(iw))**(-1)*( - vel_pz(zz,xx)**2/freq(iw)**2*(tmpdx + tmpdz)))
      !dIpr(zz,xx,rhs) = dIpr(zz,xx,rhs) + tmpdi
    enddo
   enddo
  enddo



  
  
  !t2 = MPI_Wtime()
  !if (myrank == 0) write(*,*) "The wall time for imaging is ", t2-t1

 !endif
 
 do rhs = 1,NRHS
    do xx = 1,lnx !!locTX-locHX+1
        do zz = 1,lnz !!locTZ-locHZ+1
	    YY(zz+lnz*(xx-1)+lnz*lnx*(rhs-1)) = X(zz,xx,rhs)
        end do
    end do
 end do

  write(pnum,fmt='(I2)'),myrank
  ifile=trim(rfile0)//'.'//trim(wnum)//'.'//trim(pnum)
  call complexoutb(YY,(locTZ-locHZ+1)*(locTX-locHX+1)*NRHS,ifile)

 
 call factors_dealloc( lenlocTT, locUL, locUR, locLL, locIPIV, Lpar-1, gbUL, gbUR, gbLL, gbIPIV )
 deallocate(water,waterpr,X,dX,YY,locUL, locUR, locLL, locIPIV, gbUL, gbUR, gbLL, gbIPIV )  
 
  t2 = MPI_Wtime()
  if (myrank == 0) write(*,*) "The wall time for propagation and imaging is ", t2-t1


enddo !!!%%%%%%%%%%%%%%%%%%%%%%%END LOOP OVER FREQUENCY%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!!
 

 
 
  !! Divide by source field squared per inverse scattering (unstable)
  do xx=1,lnx
   do zz = 1,lnz
    do rhs=1,NRHS/2
      !!Ipr(zz,xx,rhs) = Ipr(zz,xx,rhs)/dIpr(zz,xx,rhs)
    enddo
   enddo
  enddo
 
 
 
   headZ = locHZ
   if (headZ > 1) then
        headZ = headZ + tos
   end if
   headX = locHX
   if (headX > 1) then
       headX = headX + tos
   end if

  
  allocate(I0(Nz*Nx*NRHS/2*2))
  I0=0.
  
do rhs = 1,NRHS/2
    do xx = headX,locTX
        do zz = headZ,locTZ
	    !!WRITE OUT BOTH I and dI TO SAME VARIABLE
	    I0(zz+Nz*(xx-1)+Nz*Nx*(rhs-1)) = Ipr(zz-locHz+1,xx-locHx+1,rhs)
	    I0(zz+Nz*(xx-1)+Nz*Nx*(rhs-1)+Nz*Nx*NRHS/2) = dIpr(zz-locHz+1,xx-locHx+1,rhs)
        end do
    end do
end do

  
t1 = MPI_Wtime()
!!write(*,*) myrank,locHZ, locTZ,locHX,locTX
 call MPI_Barrier(MPI_COMM_WORLD,ierr)
 Is=0.

 call MPI_ALLREDUCE(I0,Is,Nx*Nz*NRHS/2*2,MPI_REAL,MPI_SUM,MPI_COMM_WORLD,ierr)
 !!call MPI_REDUCE(I0,Is,Nx*Nz*NRHS/2*2,MPI_REAL,MPI_SUM,0,MPI_COMM_WORLD,ierr)
 deallocate(I0)

t2 = MPI_Wtime()
 if (myrank == 0) write(*,*) "The wall time of the MPI_reduce", t2-t1


  
 if (myrank==0) then
    ifile = trim(ifile0)//'PR'
    call realoutb(Is,Nz*Nx*NRHS,ifile)
 endif
   if (myrank == 0) write(*,*) "Finished writing to file"
  
 endif  !!SKIP FORWARD PROP and GO TO ERROR FN 
  
  
  
  
  
  
  
  if (forward==0) then
    !! Read image from file if it wasn't calculated
    ifile = trim(ifile0)//'PR'
    call realinb(Is,Nz*Nx*NRHS,ifile)
  endif
   
    t1=MPI_Wtime()
    !!call find_scatters(Is, Nx, Nz, NRHS, nsi, xis, zis, si1, si2)
    
    !nsi=9016
    nsi=165652 
    !    ifile=trim(ifile0)//'x'
    !call realinb(xis,nsi,ifile)
    !    ifile=trim(ifile0)//'z'
    !call realinb(zis,nsi,ifile)
    !    ifile=trim(ifile0)//'s1'
    !call realinb(si1,nsi,ifile)
    !    ifile=trim(ifile0)//'s2'
    !call realinb(si2,nsi,ifile)
	
	if (myrank==0) write(*,*) xis(1), xis(nsi),zis(1), zis(nsi),si1(1), si2(nsi)
	
	
    allocate(Jig(nsi),Jall(nsi*ng))
    t2=MPI_Wtime()
      
      if (myrank==0) write(*,*) " "
      if (myrank==0) write(*,*) "Time for find_scatters ", t2-t1, "Error measures = ", nsi
 
    
    t1=MPI_Wtime() 
    
         
    call power_norm_adjtest3(Jig,Is,Iadj,Nx,Nz,NRHS,nsi)
    t2=MPI_Wtime()
    if (myrank==0) write(*,*) "Time for power_norm is ", t2-t1
    
    
    
    !call smootherf1D(Iadj,Nx,Nz,Nrhs)
    
    
	 !write(*,*) 'debug5',myrank
    Jsum=0.
    do ix=1,nsi
      !!Jall(ix+Nsi*(ig-1)) = Jig(ix)
      Jsum = Jsum+Jig(ix)
    enddo
    	 !if (myrank==0) write(*,*) 'debug6'
    
  if (myrank==0) then
    ifile=trim(ifile0)//'J1'
    open( unit=fps,file=TRIM(ifile), access='direct',recl=nsi,form='unformatted')
    write(unit=fps,rec=(ig))Jig
    close(fps)
    
    ifile = trim(ifile0)//'asrc'
    call realoutb(Iadj,Nz*Nx*NRHS/2,ifile)
    
    write(*,*) "Total error measurements = ",nsi
    write(*,*) "Error function = ",Jsum, "for perturbation ", ig
    write(*,*) " "
    write(*,*) " "
  endif
    
	
  
  if (myrank == 0) then
        ifile=trim(ifile0)//'x'
    !call realoutb(xis,nsi,ifile)
        ifile=trim(ifile0)//'z'
    !call realoutb(zis,nsi,ifile)
        ifile=trim(ifile0)//'s1'
    !call realoutb(si1,nsi,ifile)
        ifile=trim(ifile0)//'s2'
    !call realoutb(si2,nsi,ifile)
  endif



 !!call MPI_Barrier(MPI_COMM_WORLD,ierr)



!!MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM!!
!!MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM!!


if (adjyn==1) then

dIpr=0.
Ipr=0.

do iw=w0,wf

!! Define initial conditions

Y = ZERO
!!Y0 = ZERO

if (myrank==0) then
write(*,*) " "
write(*,*) "Adjoint propagation for frequency ",iw
write(*,*) "********************************"
endif



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  3. generating the matrix A  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
t1 = MPI_Wtime()
if (JOB == 1) then        ! isotropic acoustic
    call A2D_acoustic( freq(iw), tos, Nz, Nx, N_delta, sigma0, hz, hx, locHZ, locTZ, locHX, locTX, label, vel_pz, rho, A )
endif
!t2 = MPI_Wtime()

if (myrank == 0) then
    !write(*,*) "local matrix A generation is done! Wall time is", t2-t1
end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  4. factorization  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!t1 = MPI_Wtime()
 call factorization(A,slvl,Lpar,rkmin,tol,mr2d,flag,ctxt,nprow,npcol,myrow,mycol,mynodes, & 
                   lenlocTT,lengbTT,locTT,gbTT,locTlabel,gbTlabel,locNB,gbNB,locUL,locUR,locLL,locIPIV,gbUL,gbUR,gbLL,gbIPIV)
!t2 = MPI_Wtime()
if (myrank == 0) then
    !write(*,*) "The total factorization is done! Wall time is", t2-t1
end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  5. solution  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 lnx=locTX-locHX+1
 lnz=locTZ-locHZ+1
 allocate( X(lnz,lnx,NRHS))
 allocate(dX(lnz,lnx,NRHS))
 allocate(YY(lnz*lnx*NRHS))
   !if (myrank == 0) write(*,*) 'debuga'
 
 X = ZERO
 dX = ZERO
 
  write(wnum,fmt='(I2)')iw
  write(pnum,fmt='(I2)'),myrank
  ifile=trim(rfile0)//'.'//trim(wnum)//'.'//trim(pnum)
  call complexinb(YY,lnx*lnz*NRHS,ifile)
    !if (myrank == 0) write(*,*) 'debugb'


  do rhs = 1,NRHS/2
    do xx = 1,lnx !!locTX-locHX+1
        do zz = 1,lnz !!locTZ-locHZ+1
	   X(zz,xx,rhs) = YY(zz+lnz*(xx-1)+lnz*lnx*(rhs-1))
	   X(zz,xx,rhs+NRHS/2) = conjg(YY(zz+lnz*(xx-1)+lnz*lnx*(rhs-1+NRHS/2)))
        end do
    end do
  end do

  if (myrank==0) write(*,*) trim(ifile)
  !if (myrank == 0) write(*,*) 'debug0', maxval(real(YY),lnz*lnx*NRHS),minval(real(YY),lnz*lnx*NRHS)
    
  dIpr=0.
  do rhs = 1,NRHS/2
    do xx = headX,locTX
        do zz = headZ,locTZ
	    dIpr(zz-locHz+1,xx-locHx+1,rhs) = Iadj(zz+Nz*(xx-1)+Nz*Nx*(rhs-1))
        end do
    end do
  end do
  if (myrank == 0) write(*,*) 'debug1', maxval(Iadj,Nz*Nx*NRHS/2),minval(Iadj,Nz*Nx*NRHS/2)
  
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
   allocate(water(NRHS/2))
    ifile = trim(ifile0)//'water'
    open( unit=fps,file=TRIM(ifile), access='direct',recl=NRHS/2,form='unformatted')
    read(unit=fps,rec=(iw))water
    close(fps)
  !if (myrank == 0) write(*,*) 'debugw', water(1),water(29)

!t1 = MPI_Wtime()
  do xx=2,lnx-1
   do zz = 2,lnz-1
    do rhs=1,NRHS/2
      

      !tmpg = 0.
      !tmpg = tmpg + freq(iw)**(-2)*dIpr(zz,xx,rhs)*vel_pz(zz,xx)**2*X(zz,xx+1,rhs+NRHS/2)/(hx**2)
      !tmpg = tmpg - freq(iw)**(-2)*dIpr(zz,xx,rhs)*vel_pz(zz,xx)**2*X(zz,xx,rhs+NRHS/2)/(hx**2)
      !tmpg = tmpg - freq(iw)**(-2)*dIpr(zz,xx-1,rhs)*vel_pz(zz,xx-1)**2*X(zz,xx,rhs+NRHS/2)/(hx**2)
      !tmpg = tmpg + freq(iw)**(-2)*dIpr(zz,xx-1,rhs)*vel_pz(zz,xx-1)**2*X(zz,xx-1,rhs+NRHS/2)/(hx**2)
      !tmpg = tmpg + freq(iw)**(-2)*dIpr(zz,xx,rhs)*vel_pz(zz,xx)**2*X(zz+1,xx,rhs+NRHS/2)/(hz**2)
      !tmpg = tmpg - freq(iw)**(-2)*dIpr(zz,xx,rhs)*vel_pz(zz,xx)**2*X(zz,xx,rhs+NRHS/2)/(hz**2)
      !tmpg = tmpg - freq(iw)**(-2)*dIpr(zz-1,xx,rhs)*vel_pz(zz-1,xx)**2*X(zz,xx,rhs+NRHS/2)/(hz**2)
      !tmpg = tmpg + freq(iw)**(-2)*dIpr(zz-1,xx,rhs)*vel_pz(zz-1,xx)**2*X(zz-1,xx,rhs+NRHS/2)/(hz**2)
      

      !!tmpdi = 1/(conjg(X(zz,xx,rhs)) * X(zz,xx,rhs) + water(rhs))
      tmpdi = 1/(X(zz,xx,rhs) * X(zz,xx,rhs))  !!Not power apparently
	    
      !!!dX(zz,xx,rhs) = -(2*PI*RT*freq(iw))**(-2) * dIpr(zz,xx,rhs) * tmpdi * (X(zz,xx,rhs+NRHS/2) + tmpg)
      dX(zz,xx,rhs) = (2*PI*RT*freq(iw))**(-1) * dIpr(zz,xx,rhs) * (X(zz,xx,rhs+NRHS/2))! * tmpdi 
            
      tmpi = (X(zz,xx,rhs)) / (conjg(X(zz,xx,rhs)) * X(zz,xx,rhs) + water(rhs))
      !tmpgpx = (X(zz,xx+1,rhs)) / (conjg(X(zz,xx+1,rhs)) * X(zz,xx+1,rhs) + water(rhs))
      !tmpgmx = (X(zz,xx-1,rhs)) / (conjg(X(zz,xx-1,rhs)) * X(zz,xx-1,rhs) + water(rhs))
      !tmpgpz = (X(zz+1,xx,rhs)) / (conjg(X(zz+1,xx,rhs)) * X(zz+1,xx,rhs) + water(rhs))
      !tmpgmz = (X(zz-1,xx,rhs)) / (conjg(X(zz-1,xx,rhs)) * X(zz-1,xx,rhs) + water(rhs))
      
      !tmpg = 0.
      !tmpg = tmpg + freq(iw)**(-2)*dIpr(zz,xx,rhs)*vel_pz(zz,xx)**2*tmpgpx/(hx**2)
      !tmpg = tmpg - freq(iw)**(-2)*dIpr(zz,xx,rhs)*vel_pz(zz,xx)**2*tmpi/(hx**2)
      !tmpg = tmpg - freq(iw)**(-2)*dIpr(zz,xx-1,rhs)*vel_pz(zz,xx-1)**2*tmpi/(hx**2)
      !tmpg = tmpg + freq(iw)**(-2)*dIpr(zz,xx-1,rhs)*vel_pz(zz,xx-1)**2*tmpgmx/(hx**2)
      !tmpg = tmpg + freq(iw)**(-2)*dIpr(zz,xx,rhs)*vel_pz(zz,xx)**2*tmpgpz/(hz**2)
      !tmpg = tmpg - freq(iw)**(-2)*dIpr(zz,xx,rhs)*vel_pz(zz,xx)**2*tmpi/(hz**2)
      !tmpg = tmpg - freq(iw)**(-2)*dIpr(zz-1,xx,rhs)*vel_pz(zz-1,xx)**2*tmpi/(hz**2)
      !tmpg = tmpg + freq(iw)**(-2)*dIpr(zz-1,xx,rhs)*vel_pz(zz-1,xx)**2*tmpgmz/(hz**2)
      
      

      !!!dX(zz,xx,rhs+NRHS/2) = (2*PI*RT*freq(iw))**(-2) * dIpr(zz,xx,rhs) * (tmpi + tmpg)
      dX(zz,xx,rhs+NRHS/2) = - (2*PI*RT*freq(iw))**(-1) * dIpr(zz,xx,rhs) * (tmpi)
     
     
      !dX(zz,xx,rhs) = (2*PI*RT*freq(iw))**(-2) * dIpr(zz,xx,rhs) * conjg(X(zz,xx,rhs+NRHS/2))
      !dX(zz,xx,rhs+NRHS/2) = (2*PI*RT*freq(iw))**(-2) * dIpr(zz,xx,rhs) * (X(zz,xx,rhs))
      
     
      
      
      
      
    enddo
   enddo
  enddo
  
  deallocate(water)
  do xx=1,lnx
    do rhs=1,NRHS
      dX(1,xx,rhs) = dX(2,xx,rhs)
      dX(lnz,xx,rhs) = dX(lnz-1,xx,rhs)
    enddo  
  enddo  
   do zz = 1,lnz
    do rhs=1,NRHS
      dX(zz,1,rhs) = dX(zz,2,rhs)
      dX(zz,lnx,rhs) = dX(zz,lnx-1,rhs)
    enddo  
   enddo
 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  
  
  if (locHZ > 1) then   ! CAUTION!!
    dX(1:tos,:,:) = ZERO
  end if
  if (locHX > 1) then
    dX(:,1:tos,:) = ZERO
  end if

  !if (myrank == 0) write(*,*) 'debug3', maxval(real(dX)),minval(real(dX))
  !if (myrank == 0) write(*,*) 'debug3a', maxval(aimag(dX)),minval(aimag(dX))

!t1 = MPI_Wtime()
 call mfsolve(dX,NRHS,slvl,Lpar,flag,Nz,Nx,locHZ,locTZ,locHX,locTX,label,ctxt,nprow,npcol,myrow,mycol,mynodes, &
             lenlocTT,lengbTT,locTT,gbTT,locTlabel,gbTlabel,locNB,gbNB,locUL,locUR,locLL,locIPIV,gbUL,gbUR,gbLL,gbIPIV)
!t2 = MPI_Wtime()
if (myrank == 0) then
    !write(*,*) "The wall time of the solution to", NRHS, " RHS is ", t2-t1
end if
  if (myrank == 0) write(*,*) 'debug4', maxval(real(dX)),minval(real(dX))
  !if (myrank == 0) write(*,*) 'debug4a', maxval(aimag(dX)),minval(aimag(dX))

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!




    write(wnum,fmt='(I2)')iw
    ifile = trim(ifile0)//'Afield'//trim(wnum)
    
   headZ = locHZ
   if (headZ > 1) then
        headZ = headZ + tos
   end if
   headX = locHX
   if (headX > 1) then
       headX = headX + tos
   end if

    fieldout0=ZERO
    do rhs = 14,14
      do xx = headX,locTX
        do zz = headZ,locTZ 
	  fieldout0(zz+Nz*(xx-1))         = dX(zz-locHz+1,xx-locHx+1,14)
	  fieldout0(zz+Nz*(xx-1)+Nz*Nx*1) = dX(zz-locHz+1,xx-locHx+1,14+NRHS/2)
	  fieldout0(zz+Nz*(xx-1)+Nz*Nx*2) =  X(zz-locHz+1,xx-locHx+1,14)
	  fieldout0(zz+Nz*(xx-1)+Nz*Nx*3) =  X(zz-locHz+1,xx-locHx+1,14+NRHS/2)
        enddo
      enddo
    enddo

   call MPI_Barrier(MPI_COMM_WORLD,ierr)
   
   fieldout=ZERO
   call MPI_REDUCE(fieldout0,fieldout,Nx*Nz*4,MPI_COMPLEX,MPI_SUM,0,MPI_COMM_WORLD,ierr)
   
   !if (myrank==0) write(*,*) 'Debug',fieldout(200+Nz*300+Nz*Nx*0), fieldout(200+Nz*300+Nz*Nx*1), fieldout(200+Nz*300+Nz*Nx*2)
   if (myrank==0) call complexoutb(fieldout,Nz*Nx*4,ifile)
   if (myrank==0) write(*,*) 'Wrote out fields for src 14'



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


dvel_pz=0.

  tmpi=0.
  tmpdi=0.

  do xx=1,lnx
   do zz = 1,lnz
    do rhs=1,NRHS/2
      

      tmpi  = real(dX(zz,xx,rhs)        * -2*freq(iw)**2*vel_pz(zz,xx)**(-3)*     (X(zz,xx,rhs)) )
      tmpdi = real(dX(zz,xx,rhs+NRHS/2) * -2*freq(iw)**2*vel_pz(zz,xx)**(-3)*conjg(X(zz,xx,rhs+NRHS/2)) )
      
      !tmpi  = real(dX(zz,xx,rhs) * (X(zz,xx,rhs)) )
      !tmpdi = real(dX(zz,xx,rhs+NRHS/2) *(X(zz,xx,rhs+NRHS/2)) )  
      
      Ipr(zz,xx,rhs) = Ipr(zz,xx,rhs) + tmpi 
      Ipr(zz,xx,rhs+NRHS/2) = Ipr(zz,xx,rhs+NRHS/2) + tmpdi  
           

      !dvel_pz(zz,xx) = dvel_pz(zz,xx) + tmpi + tmpdi
      
    enddo
   enddo
  enddo

 
  t2 = MPI_Wtime()
  if (myrank == 0) write(*,*) "The wall time for adjoint propagation is ", t2-t1


 
 call factors_dealloc( lenlocTT, locUL, locUR, locLL, locIPIV, Lpar-1, gbUL, gbUR, gbLL, gbIPIV )
 deallocate(X,dX,YY,locUL, locUR, locLL, locIPIV, gbUL, gbUR, gbLL, gbIPIV )  


enddo !!!%%%%%%%%%%%%%%%%%%%%%%%END LOOP OVER FREQUENCY%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!!
 




   headZ = locHZ
   if (headZ > 1) then
        headZ = headZ + tos
   end if
   headX = locHX
   if (headX > 1) then
       headX = headX + tos
   end if


  deallocate(Is)
  !allocate(Is(Nz*Nx), I0(Nz*Nx))
  allocate(Is(Nz*Nx*NRHS), I0(Nz*Nx*NRHS))
 
  I0=0.

    do xx = headX,locTX
        do zz = headZ,locTZ
	    !!I0(zz+Nz*(xx-1)) = dvel_pz(zz-locHz+1,xx-locHx+1)
        end do
    end do
    do rhs = 1,NRHS
      do xx = headX,locTX
        do zz = headZ,locTZ
	    I0(zz+Nz*(xx-1)+Nz*Nx*(rhs-1)) = Ipr(zz-locHz+1,xx-locHx+1,rhs)
	    !!I0(zz+Nz*(xx-1)+Nz*Nx*(rhs-1)+Nz*Nx*NRHS/2) = dIpr(zz-locHz+1,xx-locHx+1,rhs)
        end do
      end do
    end do

  
t1 = MPI_Wtime()

 call MPI_Barrier(MPI_COMM_WORLD,ierr)
 
 Is=0.
 !call MPI_REDUCE(I0,Is,Nx*Nz,MPI_REAL,MPI_SUM,0,MPI_COMM_WORLD,ierr)
 call MPI_REDUCE(I0,Is,Nx*Nz*NRHS/2*2,MPI_REAL,MPI_SUM,0,MPI_COMM_WORLD,ierr)
 deallocate(I0)

t2 = MPI_Wtime()
 if (myrank == 0) write(*,*) "The wall time of the MPI_reduce", t2-t1
 

  
  if (myrank==0) then
    ifile = trim(ifile0)//'vadj'
    !call realoutb(Is,Nz*Nx,ifile)
    call realoutb(Is,Nz*Nx*NRHS,ifile)
    
  endif
  
 call MPI_Barrier(MPI_COMM_WORLD,ierr)

 if (myrank == 0) write(*,*) "Hopefully printed out"


!!MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM!!
!!MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM!!





 

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  6. deallocating factors  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 deallocate(gbTT,locTT,gbTlabel,locTlabel,label,locNB,gbNB)
 deallocate(ctxt, nprow, npcol, myrow, mycol, mynodes)
 
   
 deallocate(Y,Is,fieldout,fieldout0,vel,vel0,dv, freq, Aw, sourceX, sourceZ,dataw, taper,rayp)
 deallocate(vel_pz,dvel_pz,rho,Ipr,dIpr,ga,gz,gx)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 call MPI_Barrier(MPI_COMM_WORLD,ierr)
   endif !!END SKIP ADJOINT PROP

  
  if (myrank == 0 .and. .false.) then
        ifile=trim(ifile0)//'J'
    call realoutb(Jall,nsi*ng,ifile)
        ifile=trim(ifile0)//'x'
    call realoutb(xis,nsi,ifile)
        ifile=trim(ifile0)//'z'
    call realoutb(zis,nsi,ifile)
        ifile=trim(ifile0)//'s1'
    call realoutb(si1,nsi,ifile)
        ifile=trim(ifile0)//'s2'
    call realoutb(si2,nsi,ifile)
  endif

    deallocate(Jig,Jall)

 call MPI_Barrier(MPI_COMM_WORLD,ierr)



!! MPI Finalization
 call MPI_Finalize(ierr)



end program adjtest
