program adjtest

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!  2D massively parallel HSS multifontal Helmholtz solver: HelmPar2D  !!!!!!!!!!!!!!!
!!!!!!!!!!!           Shen Wang,  Jianlin Xia,  Maarten V. de Hoop              !!!!!!!!!!!!!!!
!!!!!!!!!!!                Purdue & ConocoPhillips, June 2010                   !!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

use NDpre2D
use context
use A2D 
use fact
use solution2D
use parser
use fileio
use powernorm
use inversion

implicit none
include 'mpif.h'


!! declaring variables
integer                      ::  ierr, myrank, nprocs
integer                      ::  JOB, tos, passive
integer                      ::  Lvl, slvl, Lpar, rkmin, mr2d,start
real(kind=4)                 ::  tol, dt, dw, step,sumbase
real(kind=8)                 ::  t1, t2
double precision             :: Jsum,Jsum0
integer                      ::  Nz, Nx,nvar, N_delta, NRHS, Nt, Nw,Nh,lnx,lnz,Np,nsi,Ng, fps=20, iter,niter,nsteps,iflag,istep,igg
 complex(kind=4)             :: sigma0, hz, hx, tmpi,tmpdi,tmpdx,tmpdz,tmpg,tmpgpx,tmpgmx,tmpgpz,tmpgmz,dPhi,tmpss
integer,allocatable          ::  sourceZ(:), sourceX(:),h(:),arrayloc(:),sources(:)
real(kind=4),allocatable     ::  Istack(:),Is(:),I0(:),Iadj(:),Ipr(:,:,:),dIpr(:,:,:)
real(kind=4),allocatable     ::  vel(:),vel0(:), dv(:), ga(:), gz(:),gx(:), taper(:), N(:),water(:),waterpr(:),rayp(:)
real,allocatable ::  vupdate(:),grad(:),grad0(:),vtmp(:),lbfgs_work(:),lbfgs_diag(:)
real(kind=4),allocatable     ::  Jig(:),Jall(:),steps(:)
 complex(kind=4),allocatable  ::  Y(:),Y0(:),YY(:), freq(:),Aw(:),dataw(:),fieldout(:),fieldout0(:)
 complex(kind=4),allocatable  ::  X(:,:,:),dX(:,:,:) 
real(kind=4)                 ::  tmp, tmpz, tmpx,alpha,maxp,xis(1000000),zis(1000000),si1(1000000),si2(1000000),betaFR,base(99,99),zp(99),xp(99)
integer                      ::  rhs, zz, xx, pp, iw, w0, wf,xi,zi, ig,ds,dh,ihh,hh, iss, ix, iz, headZ, headX,forward,adjyn
 character(len=256)           ::  vfile,v0file,dvfile,sfile,sfile0,rfile,rfile0,ifile,ifile0,wnum,pnum 
 !!complex(kind=4),parameter   :: ONE = (1.0E+0,0.0E+0), ZERO = (0.0E+0,0.0E+0), RT = (0.0E+0,1.0E+0)
real(kind=4),allocatable     ::  vel_pz(:,:), dvel_pz(:,:), rho(:,:)

  
  
!! MPI Initialization
 call MPI_Init(ierr)
 call MPI_Comm_rank( MPI_COMM_WORLD, myrank, ierr )
 call MPI_Comm_size( MPI_COMM_WORLD, nprocs, ierr )

if (myrank==0) write(*,*) "nprocs is ",nprocs

!! set up variables
 call frompar('level',Lvl,14)
 call frompar('Nz',Nz,2048)
 call frompar('Nx',Nx,2048)
 !call frompar('Ns',NRHS,1)
 call frompar('Nt',Nt,2048)
 call frompar('Nw',Nw,55)
 call frompar('Nh',Nh,20)
 call frompar('dh',dh,2)
 call frompar('dx',tmp,250.); hx=tmp
 call frompar('dz',tmp,250.); hz=tmp
 call frompar('dt',dt,0.01)
 call frompar('ds',ds,10)
 call frompar('w0',w0,2)
 call frompar('wf',wf,Nw)
 call frompar('x0',xi,80)
 call frompar('z0',zi,110)
 call frompar('alpha',alpha,1.)
 call frompar('vfile',vfile)
 call frompar('v0file',v0file)
 call frompar('dvfile',dvfile)
 call frompar('sfile',sfile0)
 call frompar('rfile',rfile0,' ')
 call frompar('ifile',ifile0)
 call frompar('passive',passive,0)
 call frompar('Ns',NRHS,29)
 call frompar('maxp',maxp,30.)
 call frompar('forward',forward,1)
 call frompar('adjoint',adjyn,1) 
 call frompar('niter',niter,1) 
 call frompar('startiter',start,1)
 call frompar('nscat',nsi)

!! choosing jobs. 1: isotropic acoustic, 2: elliptic anisotropy, 3: VTI, 4. HTI, 5. TTI
JOB = 1
tos = 1



  allocate(vel(Nz*Nx),vel0(Nz*Nx),dv(Nz*Nx))
  allocate(Jall(niter))


  vel = 0.
  vel0 = 0.
  dv = 0.
  Jall = 0.

    
 
    ig=start
    Ng=47
       
do ig=start,Ng 

NRHS = 29
  
  if (myrank==0) write(*,*) " "
  if (myrank==0) write(*,*) " "  
  if (myrank==0 .and. adjyn==1) write(*,*) "IMAGING FOR PERTURBATION ",ig
  

if (myrank==0) then
write(*,*) "Data source: ",trim(sfile0),myrank
write(*,*) "Image file:  ",trim(ifile0)
endif


if (myrank==0) write(*,*) "Number of sources", NRHS


allocate(Y(Nz*Nx*NRHS*2))
allocate(Is(Nz*Nx*NRHS*2))
allocate(Iadj(Nz*Nx*NRHS))
allocate(dataw(Nx*29))
allocate(freq(Nw),Aw(Nw))
allocate(sourceZ(NRHS),sourceX(NRHS),rayp(NRHS))         
allocate(taper(Nx),arrayloc(Nx))
allocate(ga(ng), gx(ng), gz(ng))
allocate(sources(NRHS))



dw = 2*PI/(Nt*dt)
do iw=1,Nw
  freq(iw)=(iw-1)*dw
  Aw(iw) = exp(-(real(wf+w0)/2 - real(iw))**2/(real(wf-w0)/5)**2)
enddo
sigma0 = 1.0*freq(wf)

sourceZ = 0
sourceX = 0
do rhs = 1,NRHS
    sourceZ(rhs) = zi 
    sourceX(rhs) = xi + ds*(rhs-1)
end do

!tmp = NRHS * 2 !! NOTE: TO ACCOMODATE SOURCE AND RECEIVER
NRHS = 58

    
    
    
!! read in velocity model

  call realinb(vel0,Nx*Nz,v0file)   ! Smooth starting model
  !call realinb(dv,Nx*Nz,vfile)      ! True perturbation
  
  !vel0 = vel0 + dv                  ! vel0 is now true smooth model
  
  !call realinb(dv,Nx*Nz,dvfile)     ! Perturbation for test
  
  !vel0 = vel0 - dv                  ! vel0 is true minus test perturbation
  
     
     
  if (myrank==0) write(*,*) "Loaded velocity models"

  
arrayloc = 0
!do xx=100,Nx-100,20
do xx=100,Nx-100,1
  arrayloc(xx)=1
enddo

taper=0.
do zz=1,Nx
  if (arrayloc(zz)==1) then
    do xx=1,Nx
      taper(xx)=taper(xx)+exp(-real(xx-zz)**2/2**2.)
    enddo
  endif
enddo
taper = taper/maxval(taper,Nx)

 !ifile = 'taperout'
 !call realoutb(taper,Nx,ifile)
 deallocate(arrayloc)


Is=0.


!! set up global variables
!Lvl            = 16
slvl           = 0
rkmin          = 20                                       ! minimum rank to guarantee the numerical stability
tol            = 1.0E-4
Lpar           = nint(log(real(nprocs))/log(2.0)) + 1     ! Lpar = log_2(nprocs) + 1
mr2d           = 0
N_delta        = 50





   ihh=1  !!Define original model
   gx(ihh)=160
   gz(ihh)=150
   ga(ihh)=0.
   
   do ix=160,560,50
    do iz=150,350,50
       ihh=ihh+1
       gx(ihh)=ix
       gz(ihh)=iz
       ga(ihh)=50  
     enddo
   enddo


  
  do iz=1,99
    zp(iz)=-pi/2 + (iz-1)*pi/98 
    xp(iz)=zp(iz);  
  enddo

  sumbase=0.
  do ix=1,99
     do iz=1,99
 	base(iz,ix)=COS(zp(iz))**2*COS(xp(ix))**2
	sumbase=sumbase+base(iz,ix)
     enddo
  enddo
 


 if (ig .lt. Ng) then
    if (myrank==0) write(*,*) "Running Perturbation"
    do ix=1,99
    	do iz=1,99
	  vel0(gz(ig)+iz-49 + Nz*(gx(ig)+ix-49-1)) = vel0(gz(ig)+iz-49 + Nz*(gx(ig)+ix-49-1)) - ga(ig)*base(iz,ix)
	enddo
    enddo
    
 elseif (ig == Ng) then
    
    if (myrank==0) write(*,*) "Running Checkerboard"
    
    do igg=1,Ng-2
      ga(igg)=(-1)**(igg+1)*100
      do ix=1,99
    	do iz=1,99 
	   vel0(gz(ig)+iz-49 + Nz*(gx(ig)+ix-49-1)) = vel0(gz(ig)+iz-49 + Nz*(gx(ig)+ix-49-1)) - ga(igg)*base(iz,ix)
	enddo
      enddo
    enddo
    
 endif
  
      ifile='velout3'
    call realoutb(vel0,Nx*Nz,ifile)
  
  

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  1. 2D nested dissection preprocessing  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
t1 = MPI_Wtime()
 call NDpre_generate( myrank, tos, Lvl, Lpar, Nz, Nx, lengbTT, lenlocTT, gbTT, locTT, gbTlabel, locTlabel, gbNB, locNB, locHZ, locTZ, locHX, locTX, label, flag )
t2 = MPI_Wtime()
!if (myrank == 0) then
    !write(*,*) "2D nested dissection domain decomposition is done! wall time is ", t2-t1
!end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  2. BLACS contexts generating  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 call ctxt_generate( flag, lengbTT, gbTT, ctxt, nprow, npcol, myrow, mycol, mynodes ) 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!! find vel model for processor
    allocate(vel_pz(locTZ-locHZ+1,locTX-locHX+1))
    allocate(dvel_pz(locTZ-locHZ+1,locTX-locHX+1))
    do xx = locHX,locTX
        do zz = locHZ,locTZ
	    vel_pz(zz-locHZ+1,xx-locHX+1) = vel0(zz+Nz*(xx-1))  !starting model
        end do
    end do
  
  allocate(rho(locTZ-locHZ+1+2*tos,locTX-locHX+1+2*tos))
  rho = 1.0E+0


  allocate(Ipr(locTZ-locHZ+1,locTX-locHX+1,NRHS))
  Ipr=0.
  allocate(dIpr(locTZ-locHZ+1,locTX-locHX+1,NRHS))
  dIpr=0.






do iw=w0,wf

!! Define initial conditions

Y = ZERO
!!Y0 = ZERO

!if (myrank==0) then
!write(*,*) " "
!write(*,*) "Propagating for frequency ",iw
!write(*,*) "********************************"
!endif



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  3. generating the matrix A  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
t1 = MPI_Wtime()
if (JOB == 1) then        ! isotropic acoustic
    call A2D_acoustic( freq(iw), tos, Nz, Nx, N_delta, sigma0, hz, hx, locHZ, locTZ, locHX, locTX, label, vel_pz, rho, A )
endif
t2 = MPI_Wtime()

!if (myrank == 0) then
    !write(*,*) "local matrix A generation is done! Wall time is", t2-t1
!end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  4. factorization  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!t1 = MPI_Wtime()
 call factorization(A,slvl,Lpar,rkmin,tol,mr2d,flag,ctxt,nprow,npcol,myrow,mycol,mynodes, & 
                   lenlocTT,lengbTT,locTT,gbTT,locTlabel,gbTlabel,locNB,gbNB,locUL,locUR,locLL,locIPIV,gbUL,gbUR,gbLL,gbIPIV)
!t2 = MPI_Wtime()
if (myrank == 0) then
  ! write(*,*) "The total factorization is done! Wall time is", t2-t1
end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  4A. define initial wavefields !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  write(wnum,fmt='(I2)')iw
  sfile=trim(sfile0)//trim(wnum)
  call complexinb(dataw,Nx*NRHS/2,sfile)

  do rhs=1,NRHS/2
    do xx=1,Nx
      do zz=2,Nz-1
         tmpz = real((zz-sourceZ(rhs))**2)
	 tmpx = real((xx-sourceX(rhs))**2)
	 if (tmpz<=16.0) then
	                                                                                
            Y(sourceZ(rhs)+1+Nz*(xx-1)+Nz*Nx*(rhs-1+NRHS/2))= Aw(iw)*taper(xx)*exp(-tmpz/16.0)*conjg(dataw((xx-1)+Nx*(rhs-1)))
            Y(sourceZ(rhs)-1+Nz*(xx-1)+Nz*Nx*(rhs-1+NRHS/2))=-Aw(iw)*taper(xx)*exp(-tmpz/16.0)*conjg(dataw((xx-1)+Nx*(rhs-1)))	    
            !Y(sourceZ(rhs)+Nz*(xx-1)+Nz*Nx*(rhs-1+NRHS/2))=Aw(iw)*taper(xx)*exp(-tmpz/16.0)*freq(iw)*conjg(dataw((xx-1)+Nx*(rhs-1)))
	 endif
	enddo
    enddo
  enddo 
    !if (myrank==0) write(*,*) "debug5",maxval(abs(Y),Nz*Nx*NRHS)
if (passive == 1) then
  write(wnum,fmt='(I2)')iw
  sfile=trim(sfile0)//'s'//trim(wnum)
  call complexinb(dataw,Nx*NRHS/2,sfile)
  do rhs=1,NRHS/2
    do xx=1,Nx
      do zz=2,Nz-1
         tmpz = real((zz-sourceZ(rhs))**2)
	 tmpx = real((xx-sourceX(rhs))**2)
         if (tmpz <= 25.0E+0) then
            !Y(zz+Nz*(xx-1)+Nz*Nx*(rhs-1)) = (dataw((xx-1)+Nx*(rhs-1)))*exp(-(tmpz)/4.0)!*Aw(iw)
	    Y(zz+Nz*(xx-1)+Nz*Nx*(rhs-1)) = (dataw((xx-1)+Nx*(rhs-1)))*exp(-(tmpz)/4.0)!*Aw(iw)
         end if
      enddo
    enddo
  enddo
	 
else 
   
  do rhs=1,NRHS/2
    do xx=1,Nx
      do zz=2,Nz-1
         tmpz = real((zz-sourceZ(rhs))**2)
	 tmpx = real((xx-sourceX(rhs))**2)
         if (tmpx+tmpz <= 25.0E+0) then
	    Y(zz+Nz*(xx-1)+Nz*Nx*(rhs-1)) = Aw(iw)*exp(-(tmpx+tmpz)/4.0)	    
         end if
	enddo
    enddo
  enddo
endif

	
  Y=Y*100 !AVOID STUPID PROBLEMS 
	 

!write(*,*) "debug6",maxval(abs(Y),Nz*Nx*NRHS)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  5. solution  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
lnx=locTX-locHX+1
lnz=locTZ-locHZ+1
allocate( X(lnz,lnx,NRHS))
allocate(dX(lnz,lnx,NRHS))
allocate(YY(lnz*lnx*NRHS))

 !write(*,*) "debuga"
X = ZERO
dX = ZERO
do rhs = 1,NRHS
    do xx = locHX,locTX
        do zz = locHZ,locTZ
	    X(zz-locHz+1,xx-locHx+1,rhs) = Y(zz+Nz*(xx-1)+Nz*Nx*(rhs-1))  !!Y(zz,xx,rhs)
	    dX(zz-locHz+1,xx-locHx+1,rhs) = Y(zz+Nz*(xx-1)+Nz*Nx*(rhs-1))  !! Compute unperturbed field
	    
        end do
    end do
end do
if (locHZ > 1) then   ! CAUTION!!
    X(1:tos,:,:) = ZERO
end if
if (locHX > 1) then
    X(:,1:tos,:) = ZERO
end if

 !write(*,*) "debugb"
 
!t1 = MPI_Wtime()
 call mfsolve(X,NRHS,slvl,Lpar,flag,Nz,Nx,locHZ,locTZ,locHX,locTX,label,ctxt,nprow,npcol,myrow,mycol,mynodes, &
             lenlocTT,lengbTT,locTT,gbTT,locTlabel,gbTlabel,locNB,gbNB,locUL,locUR,locLL,locIPIV,gbUL,gbUR,gbLL,gbIPIV)
!t2 = MPI_Wtime()
if (myrank == 0) then
    !write(*,*) "The wall time of the solution to", NRHS, " RHS is ", t2-t1
end if


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 !write(*,*) "debugc"

 allocate(waterpr(NRHS/2),water(NRHS/2))
 waterpr = 0.
 water = 0.
 do xx=1,locTX-locHX+1
   do zz = 1,locTZ-locHZ+1
      do rhs=1,NRHS/2
        tmp = 0.2*real(X(zz,xx,rhs)*conjg(X(zz,xx,rhs)))
        if (waterpr(rhs) .lt. tmp) then
           waterpr(rhs)=tmp
        endif
      enddo
   enddo
 enddo
 
 call MPI_ALLREDUCE(waterpr,water,NRHS/2,MPI_REAL,MPI_MAX,MPI_COMM_WORLD,ierr)
 if (myrank==0) then
    ifile = trim(ifile0)//'water'
    open( unit=fps,file=TRIM(ifile), access='direct',recl=NRHS/2,form='unformatted')
    write(unit=fps,rec=(iw))water
    close(fps)
 endif

  

  do xx=1,lnx
   do zz = 1,lnz
    do rhs=1,NRHS/2
      
      
      !tmpg = (X(zz,xx,rhs)) / (conjg(X(zz,xx,rhs)) * X(zz,xx,rhs) + water(rhs))
      !tmpgpx = (X(zz,xx+1,rhs)) / (conjg(X(zz,xx+1,rhs)) * X(zz,xx+1,rhs) + water(rhs))
      !tmpgmx = (X(zz,xx-1,rhs)) / (conjg(X(zz,xx-1,rhs)) * X(zz,xx-1,rhs) + water(rhs))
      !tmpgpz = (X(zz+1,xx,rhs)) / (conjg(X(zz+1,xx,rhs)) * X(zz+1,xx,rhs) + water(rhs))
      !tmpgmz = (X(zz-1,xx,rhs)) / (conjg(X(zz-1,xx,rhs)) * X(zz-1,xx,rhs) + water(rhs))
      tmpg = (X(zz,xx,rhs))     !/ (conjg(X(zz,xx,rhs)) * X(zz,xx,rhs) + water(rhs))
      tmpgpx = (X(zz,xx+1,rhs)) !/ (conjg(X(zz,xx+1,rhs)) * X(zz,xx+1,rhs) + water(rhs))
      tmpgmx = (X(zz,xx-1,rhs)) !/ (conjg(X(zz,xx-1,rhs)) * X(zz,xx-1,rhs) + water(rhs))
      tmpgpz = (X(zz+1,xx,rhs)) !/ (conjg(X(zz+1,xx,rhs)) * X(zz+1,xx,rhs) + water(rhs))
      tmpgmz = (X(zz-1,xx,rhs)) !/ (conjg(X(zz-1,xx,rhs)) * X(zz-1,xx,rhs) + water(rhs))
      

      tmpi = (X(zz,xx,rhs+NRHS/2)) *  tmpg !!(X(zz,xx,rhs)) !tmpg
      tmpdi = (X(zz,xx,rhs+NRHS/2)) *  (X(zz,xx,rhs)) !tmpg
      !tmpdi = (conjg(X(zz,xx,rhs)) * X(zz,xx,rhs) + water(rhs))
	    
      if (xx.gt.1 .and. xx.lt.lnx) then
        tmpdx = (tmpgpx-tmpgmx)*(X(zz,xx+1,rhs+NRHS/2)-X(zz,xx-1,rhs+NRHS/2))/(4*hx**2) 
      else if (xx.eq.1) then
        tmpdx = (tmpgpx-tmpg)*(X(zz,xx+1,rhs+NRHS/2)-X(zz,xx,rhs+NRHS/2))/(hx**2) 
      else if (xx.eq.lnx)  then      
        tmpdx = (tmpg-tmpgmx)*(X(zz,xx,rhs+NRHS/2)-X(zz,xx-1,rhs+NRHS/2))/(hx**2) 
      endif
      if (zz.gt.1 .and. zz.lt.lnz) then
        tmpdz = (tmpgpz-tmpgmz)*(X(zz+1,xx,rhs+NRHS/2)-X(zz-1,xx,rhs+NRHS/2))/(4*hz**2)
      else if (zz.eq.1) then
        tmpdz = (tmpgpz-tmpg)*(X(zz+1,xx,rhs+NRHS/2)-X(zz,xx,rhs+NRHS/2))/(hz**2) 
      else if (zz.eq.lnz)  then      
        tmpdz = (tmpg-tmpgmz)*(X(zz,xx,rhs+NRHS/2)-X(zz-1,xx,rhs+NRHS/2))/(hz**2) 
      endif
     
      Ipr(zz,xx,rhs) = Ipr(zz,xx,rhs) + freq(iw)*real((PI*2*RT*freq(iw))**(-1)*(tmpdi - vel_pz(zz,xx)**2/freq(iw)**2*(tmpdx + tmpdz)))
      !Ipr(zz,xx,rhs) = Ipr(zz,xx,rhs) + real((PI*2*RT*freq(iw))**(-1)*(tmpi - vel_pz(zz,xx)**2/freq(iw)**2*(tmpdx + tmpdz)))
      !Ipr(zz,xx,rhs) = Ipr(zz,xx,rhs) + real(tmpdi)


    enddo
   enddo
  enddo


 do rhs = 1,NRHS
    do xx = 1,lnx !!locTX-locHX+1
        do zz = 1,lnz !!locTZ-locHZ+1
	    YY(zz+lnz*(xx-1)+lnz*lnx*(rhs-1)) = X(zz,xx,rhs)
        end do
    end do
 end do

 
 call factors_dealloc( lenlocTT, locUL, locUR, locLL, locIPIV, Lpar-1, gbUL, gbUR, gbLL, gbIPIV )
 deallocate(water,waterpr,X,dX,YY,locUL, locUR, locLL, locIPIV, gbUL, gbUR, gbLL, gbIPIV )  
 
  t2 = MPI_Wtime()
  if (myrank == 0 .and. mod(iw,20).eq.0) write(*,*) "Frequency",iw," Wall time for imaging ", t2-t1

enddo !!!%%%%%%%%%%%%%%%%%%%%%%%END LOOP OVER FREQUENCY%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!!
 

 
 
 
   headZ = locHZ
   if (headZ > 1) then
        headZ = headZ + tos
   end if
   headX = locHX
   if (headX > 1) then
       headX = headX + tos
   end if

  
  allocate(I0(Nz*Nx*NRHS/2*2))
  I0=0.
  
do rhs = 1,NRHS/2
    do xx = headX,locTX
        do zz = headZ,locTZ
	    !!WRITE OUT BOTH I and dI TO SAME VARIABLE
	    I0(zz+Nz*(xx-1)+Nz*Nx*(rhs-1)) = Ipr(zz-locHz+1,xx-locHx+1,rhs)
	    I0(zz+Nz*(xx-1)+Nz*Nx*(rhs-1)+Nz*Nx*NRHS/2) = dIpr(zz-locHz+1,xx-locHx+1,rhs)
        end do
    end do
end do

  
 t1 = MPI_Wtime()

 call MPI_Barrier(MPI_COMM_WORLD,ierr)
 Is=0.

 call MPI_ALLREDUCE(I0,Is,Nx*Nz*NRHS/2*2,MPI_REAL,MPI_SUM,MPI_COMM_WORLD,ierr)
 deallocate(I0)

 t2 = MPI_Wtime()
 !if (myrank == 0) write(*,*) "The wall time of the MPI_reduce", t2-t1


  
 if (myrank==0) then
    write(pnum,fmt='(I2)'),ig
    ifile = trim(ifile0)//'PR'//trim(pnum)
    !call realoutb(Is,Nz*Nx*NRHS,ifile)
 endif


  
  

    !nsi=9016
    !nsi=165652 
    allocate(Jig(nsi))
    
    t1=MPI_Wtime()          
    call power_norm_adjtest3(Jig,Is,Iadj,Nx,Nz,NRHS,nsi)
    t2=MPI_Wtime()
    
    if (myrank==0) write(*,*) "Time for power_norm is ", t2-t1
    
    Jsum=0.
    do ix=1,nsi
      Jsum = Jsum+Jig(ix)
    enddo
    Jall(iter) = Jsum
    
  if (myrank==0) then
    write(pnum,fmt='(I2)'),ig  
    ifile=trim(ifile0)//'J'//trim(pnum)
    call realoutb(Jig,nsi,ifile)
    
    
    write(*,*) "Total error measurements = ",nsi
    !write(*,*) " "   
    write(*,*) "Error function = ",Jsum, "for perturbation ", ig
    !write(*,*) " "
    write(*,*) " "
  endif
    



 call MPI_Barrier(MPI_COMM_WORLD,ierr)




    deallocate(Y)
    !write(*,*) "Debug1"
    deallocate(ga)
     !write(*,*) "Debug2a"   
    deallocate(Is)
     !   write(*,*) "Debug2"
    deallocate(Iadj)
     !  write(*,*) "Debug3" 
    !deallocate(vel0)
     !   write(*,*) "Debug4"
    !deallocate(dv)
     !   write(*,*) "Debug5"
    deallocate(freq)
     !   write(*,*) "Debug6"
    deallocate(Aw)
     !   write(*,*) "Debug7"
    deallocate(sourceX)
     !   write(*,*) "Debug8"
    deallocate(sourceZ)
     !   write(*,*) "Debug8"
    deallocate(  dataw) 
     !    write(*,*) "Debug9"
    deallocate( taper)
     !write(*,*) "Debug10"   
    deallocate(rayp)
     ! write(*,*) "Debug11"  
    deallocate( gz )  
     !   write(*,*) "Debug12"
    deallocate(gx)
     !   write(*,*) "Debug13"
    deallocate(Jig)
     !   write(*,*) "Debug41"
    deallocate(sources)
     !   write(*,*) "Debug15"
	
    


 deallocate(gbTT,locTT,gbTlabel,locTlabel,label,locNB,gbNB)
 deallocate(ctxt, nprow, npcol, myrow, mycol, mynodes)
 
 deallocate(vel_pz,dvel_pz,Ipr,dIpr,rho)



    
  enddo  !!END LOOP 

  
  
    deallocate(vel,Jall,vel0,dv)

 call MPI_Barrier(MPI_COMM_WORLD,ierr)



!! MPI Finalization
 call MPI_Finalize(ierr)



end program adjtest
