program migration

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!  2D massively parallel HSS multifontal Helmholtz solver: HelmPar2D  !!!!!!!!!!!!!!!
!!!!!!!!!!!           Shen Wang,  Jianlin Xia,  Maarten V. de Hoop              !!!!!!!!!!!!!!!
!!!!!!!!!!!                Purdue & ConocoPhillips, June 2010                   !!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


use parser
!!use propagate_w

implicit none
include 'mpif.h'


!! declaring variables
integer                      ::  ierr, myrank, nprocs
integer                      ::  JOB, tos
integer                      ::  Lvl, slvl, Lpar, rkmin, mr2d
real(kind=4)                 ::  tol, dt, dw
real(kind=8)                 ::  t1, t2
integer                      ::  Nz, Nx, N_delta, NRHS, Nt, Nw,Nh
 complex(kind=4)              :: sigma0, hz, hx, tmpi,tmpdx,tmpdz
integer,allocatable          ::  sourceZ(:), sourceX(:),h(:)
real(kind=4),allocatable     ::  I(:),Is(:),I0(:),Ih(:),vel(:),vel0(:), dv(:), ga(:), gz(:),gx(:), taper(:), N(:)
 complex(kind=4),allocatable  ::  X(:),X0(:), freq(:),Aw(:),dataw(:)
real(kind=4)                 ::  tmp, tmpz, tmpx,alpha
integer                      ::  rhs, zz, xx, pp, iw, w0, wf,xi,zi, ig,ds,dh,ihh,hh, iss, ix, iz
 character(len=256)           ::  vfile,v0file,dvfile,sfile,sfile0,rfile,rfile0,ifile,ifile0,wnum 
 complex(kind=4),parameter   :: pi = (3.1415926E+0,0.0E+0)
 complex(kind=4),parameter   :: ONE = (1.0E+0,0.0E+0), ZERO = (0.0E+0,0.0E+0), RT = (0.0E+0,1.0E+0)


!! MPI Initialization
 call MPI_Init(ierr)
 call MPI_Comm_rank( MPI_COMM_WORLD, myrank, ierr )
 call MPI_Comm_size( MPI_COMM_WORLD, nprocs, ierr )


!! set up variables
 call frompar('level',Lvl,10)
 call frompar('Nz',Nz,2048)
 call frompar('Nx',Nx,2048)
 !call frompar('Ns',NRHS,1)
 call frompar('Nt',Nt,2048)
 call frompar('Nw',Nw,55)
 call frompar('Nh',Nh,20)
 call frompar('dh',dh,2)
 call frompar('dx',tmp,250.); hx=tmp
 call frompar('dz',tmp,250.); hz=tmp
 call frompar('dt',dt,0.01)
 call frompar('ds',ds,10)
 call frompar('w0',w0,2)
 call frompar('wf',wf,Nw)
 call frompar('x0',xi,80)
 call frompar('z0',zi,110)
 call frompar('alpha',alpha,1.)
 call frompar('vfile',vfile)
 call frompar('v0file',v0file)
 call frompar('dvfile',dvfile)
 call frompar('sfile',sfile0)
 call frompar('rfile',rfile0,' ')
 call frompar('ifile',ifile0)

if (myrank==0) then
write(*,*) "Data source: ",trim(sfile0)
write(*,*) "Image file:  ",trim(ifile0)
endif
!vfile      = "/autohome/u111/wang273/data/BP2007/Vp_Model.dat"
!sfile0      = "~/Hsolver/f90/data/stest"


NRHS=int((Nx-2*xi)/ds + 1)
if (myrank==0) then
write(*,*) "Number of sources", NRHS
endif



allocate(X(Nz*Nx*NRHS*2),X0(Nz*Nx*NRHS))
allocate(I(Nz*Nx))
allocate(Is(Nz*Nx*NRHS))
!allocate(Ih(Nz*Nx*(2*Nh-1)))
allocate(I0(Nz*Nx*NRHS))
allocate(N(Nz*Nx*NRHS))
allocate(dataw(Nx*NRHS))
allocate(freq(Nw),Aw(Nw))
allocate(sourceZ(NRHS),sourceX(NRHS))
allocate(vel(Nz*Nx))
allocate(vel0(Nz*Nx))
allocate(dv(Nz*Nx))
allocate(taper(Nx))
allocate(ga(8), gx(8), gz(8))
allocate(h(2*Nh-1))

dw = 2*PI/(Nt*dt)
do iw=1,Nw
  freq(iw)=(iw-1)*dw
  Aw(iw) = exp(-(real(wf+w0)/2 - real(iw))**2/(real(wf-w0)/5)**2)
enddo
sigma0 = 1.0*freq(wf)

sourceZ = 0
sourceX = 0
do rhs = 1,NRHS
    sourceZ(rhs) = zi 
    sourceX(rhs) = xi + ds*(rhs-1)
end do

do ihh=1,2*Nh-1
  h(ihh)=dh*(ihh-Nh)
enddo

NRHS = NRHS * 2 !! NOTE: TO ACCOMODATE SOURCE AND RECEIVER

!! read in velocity model

if (1 > 0) then
  call realinb(vel,Nx*Nz,vfile)
  call realinb(vel0,Nx*Nz,v0file)
  !call realinb(dv,Nx*Nz,dvfile)
  do xx=1,Nx*Nz
    vel(xx)=vel0(xx)+alpha*vel(xx)
  enddo
  if (myrank==0) then
  write(*,*) "Loaded velocity models"
  endif
else
    gx=(/350., 450., 550., 650., 350., 450., 550., 650./)
    gz=(/350., 350., 350., 350., 550., 550., 550., 500./)
    ga=(/1., 2., 2., 1., 2., 3., 3., 2./) 
    vel = 7200.0E+0
    !!dv = 0.0
    do xx=1,Nx
    	do zz=1,Nz
	do ig =1,8
	   tmp = 100*ga(ig)*real((zz-gz(ig))**2+(xx-gx(ig))**2)
	   vel(zz+Nz*(xx-1)) = vel(zz+Nz*(xx-1)) + exp(-tmp/2.)
	enddo
	enddo

    enddo
endif
if (myrank==0) then
write(*,*) "debug4"
endif

taper=0.
do zz=201,Nx-200
  do xx=1,Nx
    taper(xx)=taper(xx)+exp(-real(xx-zz)**2/50**2.)
  enddo
enddo
taper = taper/maxval(taper,Nx)

I=0.
Is=0.
I0=0.
!Ih=0.

do iw=w0,wf

!! Define initial conditions

X = ZERO
X0 = ZERO

if (myrank==0) then
write(*,*) " "
write(*,*) "Propagating for frequency ",iw
write(*,*) "********************************"
endif

  write(wnum,fmt='(I2)')iw
  sfile=trim(sfile0)//trim(wnum)
  call complexinb(dataw,Nx*NRHS/2,sfile)


  do rhs=1,NRHS/2
    do xx=1,Nx
      do zz=2,Nz-1
         tmpz = real((zz-sourceZ(rhs))**2)
	 tmpx = real((xx-sourceX(rhs))**2)
	 if (tmpz<=16.0) then
            !X(sourceZ(rhs)+1+Nz*(xx-1)+Nz*Nx*(rhs-1+NRHS/2))=Aw(iw)*taper(xx)*exp(-tmpz/16.0)*conjg(dataw((xx-1)+Nx*(rhs-1)))
            !X(sourceZ(rhs)-1+Nz*(xx-1)+Nz*Nx*(rhs-1+NRHS/2))=-Aw(iw)*taper(xx)*exp(-tmpz/16.0)*conjg(dataw((xx-1)+Nx*(rhs-1)))	    
            X(sourceZ(rhs)+Nz*(xx-1)+Nz*Nx*(rhs-1+NRHS/2))=Aw(iw)*taper(xx)*exp(-tmpz/16.0)*conjg(dataw((xx-1)+Nx*(rhs-1)))
	 endif
         if (tmpx+tmpz <= 25.0E+0) then
	    X(zz+Nz*(xx-1)+Nz*Nx*(rhs-1)) = Aw(iw)*exp(-(tmpx+tmpz)/4.0)	    
         end if
      enddo
    enddo
  enddo
  if (myrank==0) then
  !write(*,*) "debug6"
  endif

 call propagate_freq(freq(iw),Nx,Nz,NRHS,hx,hz,X,vel,Lvl,sigma0,myrank,nprocs)
 !call propagate_freq(freq(iw),Nx,Nz,NRHS,hx,hz,X0,vel,Lvl,sigma0,myrank,nprocs)

 if (myrank == 0) then
  
  
  do ix=101,Nx-100 !2,Nx-1
   do zz = 2,Nz-1
    do iss=1,NRHS/2
      xx=Nz*(ix-1)
      rhs = Nz*Nx*(iss-1)
      !Is(zz+xx+rhs) = Is(zz+xx+rhs) + Aw(iw)**2*real(X(zz+xx+rhs)*X(zz+xx+rhs+Nz*Nx*NRHS/2))
      !I(zz+xx) = I(zz+xx) + Aw(iw)**2*real(X(zz+xx+rhs) * X(zz+xx+rhs+Nz*Nx*NRHS/2))
      
      tmpi = X(zz+xx+rhs) * X(zz+xx+rhs+Nz*Nx*NRHS/2)
      tmpdx = (X(zz+1*Nz+xx+rhs)-X(zz-1*Nz+xx+rhs))*(X(zz+1*Nz+xx+rhs+Nz*Nx*NRHS/2)-X(zz-1*Nz+xx+rhs+Nz*Nx*NRHS/2))/(4*hx**2) 
      tmpdz = (X(zz+1+xx+rhs)-X(zz-1+xx+rhs))*(X(zz+1+xx+rhs+Nz*Nx*NRHS/2)-X(zz-1+xx+rhs+Nz*Nx*NRHS/2))/(4*hz**2)
      
      I(zz+xx) = I(zz+xx) + real(PI**(-1)*2*RT*freq(iw)**(-3)*(freq(iw)**2*vel(zz+xx)**(-2)*tmpi - tmpdx - tmpdz))
      Is(zz+xx+rhs) = Is(zz+xx+rhs) + real(PI**(-1)*2*RT*freq(iw)**(-3)*(freq(iw)**2*vel(zz+xx)**(-2)*tmpi - tmpdx - tmpdz))
      !I0(zz+xx+rhs) = I0(zz+xx+rhs) + real(tmpi)
      N(zz+xx+rhs) = N(zz+xx+rhs) + real(X(zz+xx+rhs)*conjg(X(zz+xx+rhs)))
      !do ihh=1,2*Nh-1
        !hh=Nz*Nx*(ihh-1)
        !tmpi = X(zz+xx+Nz*h(ihh)+rhs) * X(zz+xx-Nz*h(ihh)+rhs+Nz*Nx*NRHS/2)
        !tmpdx = (X(zz+1*Nz+xx+Nz*h(ihh)+rhs)-X(zz-1*Nz+xx+Nz*h(ihh)+rhs))*(X(zz+1*Nz+xx-Nz*h(ihh)+rhs+Nz*Nx*NRHS/2)-X(zz-1*Nz+xx-Nz*h(ihh)+rhs+Nz*Nx*NRHS/2))/(4*hx**2) 
        !tmpdz = (X(zz+1+xx+Nz*h(ihh)+rhs)-X(zz-1+xx+Nz*h(ihh)+rhs))*(X(zz+1+xx-Nz*h(ihh)+rhs+Nz*Nx*NRHS/2)-X(zz-1+xx-Nz*h(ihh)+rhs+Nz*Nx*NRHS/2))/(4*hz**2)
        !Ih(zz+xx+hh) = Ih(zz+xx+hh) + real(PI**(-1)*2*RT*freq(iw)**(-3)*(freq(iw)**2*vel(zz+xx)**(-2)*tmpi - tmpdx - tmpdz))
        !!Ih(zz+xx+hh) = Ih(zz+xx+hh) + real(X(zz+xx+Nz*h(ihh)+rhs)*X(zz+xx-Nz*h(ihh)+rhs+Nz*Nx*NRHS/2))	
      !enddo

    enddo
   enddo
  enddo
  write(wnum,fmt='(I2)')iw
  !!ifile = trim(ifile0)//trim(wnum)
  !call realoutb(I,Nz*Nx,ifile0)
  !!call realoutb(I,Nz*Nx,ifile)
  ifile = trim(ifile0)//'S'
  call realoutb(Is,Nz*Nx*NRHS/2,ifile)
  !ifile = trim(ifile0)//'H'
  !call realoutb(Ih,Nz*Nx*(2*Nh-1),ifile)
  !call realoutb(N,Nz*Nx*NRHS/2,rfile0)
  
  !rfile=trim(rfile0)//trim(wnum)
  !call complexoutb(X(Nx*Nz*(20)),Nz*Nx,rfile)
  !rfile=trim(rfile0)//'r'//trim(wnum)
  !call complexoutb(X(Nx*Nz*(NRHS/2+20)),Nz*Nx,rfile)
 endif
 
enddo

  if (myrank==0) then
    ifile = trim(ifile0)//'S'
    call realoutb(I,Nz*Nx,ifile0)
    call realoutb(Is/N,Nz*Nx*NRHS/2,ifile)
    !ifile = trim(ifile0)//'0'
    !call realoutb(I0,Nz*Nx*NRHS/2,ifile)
    !!call realoutb(N,Nz*Nx*NRHS/2,rfile0)
    !!ifile=trim(ifile0)//'H'
    !!call realoutb(Ih,Nz*Nx*(2*Nh-1)) 
 endif

deallocate(X,X0,I,Is,I0,N, vel,vel0,dv, freq, Aw, sourceX, sourceZ,h,dataw, taper)
!deallocate(Ih)


!! MPI Finalization
 call MPI_Finalize(ierr)



end program migration
