program modeling

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!  2D massively parallel HSS multifontal Helmholtz solver: HelmPar2D  !!!!!!!!!!!!!!!
!!!!!!!!!!!           Shen Wang,  Jianlin Xia,  Maarten V. de Hoop              !!!!!!!!!!!!!!!
!!!!!!!!!!!                Purdue & ConocoPhillips, June 2010                   !!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



use NDpre2D
use context
use A2D 
use fact
use solution2D
use parser
use fileio
use powernorm
use inversion


implicit none
include 'mpif.h'


!! declaring variables
integer                      ::  ierr, myrank, nprocs
integer                      ::  JOB, tos,passive
integer                      ::  Lvl, slvl, Lpar, rkmin, mr2d
real(kind=4)                 ::  tol, dt, dw
real(kind=8)                 ::  t1, t2
integer                      ::  Nz, Nx, N_delta, NRHS, Nt, Nw,Np,lnx,lnz
 complex(kind=4)              :: sigma0, hz, hx
integer,allocatable          ::  sourceZ(:), sourceX(:)
 real(kind=4),allocatable     ::  vel(:),vel0(:), dv(:), ga(:), gz(:),gx(:),rayp(:), taper(:)
  complex(kind=4),allocatable  ::  Y(:),Y0(:),YY(:), freq(:),Aw(:),dataw(:) 
  complex(kind=4),allocatable  ::  X(:,:,:),X0(:,:,:) 
 real(kind=4)                 ::  tmp,maxp
 integer                      ::  rhs, zz, xx, pp, iw, w0, wf,xi,zi, ig,ds,dh,ihh,hh, iss, ix, iz, headZ, headX,forward,adjyn
 character(len=256)           ::  vfile,v0file,dvfile,sfile,sfile0,rfile,rfile0,wnum 
 !complex(kind=4),parameter   :: pi = (3.1415926E+0,0.0E+0)
 !complex(kind=4),parameter   :: ONE = (1.0E+0,0.0E+0), ZERO = (0.0E+0,0.0E+0), RT = (0.0E+0,1.0E+0)
 real(kind=4),allocatable     ::  vel_pz(:,:), dvel_pz(:,:), rho(:,:)







  



!! MPI Initialization
 call MPI_Init(ierr)
 call MPI_Comm_rank( MPI_COMM_WORLD, myrank, ierr )
 call MPI_Comm_size( MPI_COMM_WORLD, nprocs, ierr )


!! set up variables
 call frompar('level',Lvl,12)
 call frompar('Nz',Nz,2048)
 call frompar('Nx',Nx,2048)
 !call frompar('Ns',NRHS,1)
 call frompar('Nt',Nt,2048)
 call frompar('Nw',Nw,55)
 call frompar('dx',tmp,250.); hx=tmp
 call frompar('dz',tmp,250.); hz=tmp
 call frompar('dt',dt,0.01)
 call frompar('ds',ds,10)
 call frompar('x0',xi,50)
 call frompar('z0',zi,50)
 call frompar('w0',w0,2)
 call frompar('wf',wf,Nw)
 call frompar('vfile',vfile)
 call frompar('dvfile',dvfile)
 call frompar('v0file',v0file)
 call frompar('sfile',sfile0)
 call frompar('rfile',rfile0,' ')
 call frompar('passive',passive,0)
 call frompar('slownesses',Np,15)
 call frompar('maxp',maxp,30.)
 
if (myrank==0) then
write(*,*) "debug2",trim(sfile0)
endif
!vfile      = "/autohome/u111/wang273/data/BP2007/Vp_Model.dat"
!sfile0      = "~/Hsolver/f90/data/stest"


if (passive == 1) then
NRHS = Np*2 - 1
else
NRHS=int((Nx-2*xi)/ds + 1)
endif

if (myrank==0) then
write(*,*) "Number of sources", NRHS
endif


allocate(Y(Nz*Nx*NRHS))
allocate(Y0(Nz*Nx*NRHS))
allocate(YY(Nz*NX*NRHS))
allocate(X(Nz,Nx,NRHS),X0(Nz,Nx,NRHS))
allocate(dataw(Nx*NRHS))
allocate(freq(Nw),taper(Nx))
allocate(sourceZ(NRHS),sourceX(NRHS),rayp(NRHS))
allocate(vel(Nz*Nx))
allocate(dv(Nz*Nx))
allocate(vel0(Nz*Nx))
allocate(ga(8), gx(8), gz(8))

dw = 2*PI/(Nt*dt)
do iw=1,Nw
  freq(iw)=(iw-1)*dw
enddo
sigma0 = 1.0*freq(wf)
if (myrank==0) then
!write(*,*) "debug3",freq(Nw)
endif



!! read in velocity model


  call realinb(vel,Nx*Nz,vfile)
  call realinb(vel0,Nx*Nz,v0file)
  call realinb(dv,Nx*Nz,dvfile)

  vel0=vel+vel0
  vel=vel0+dv

  if (myrank==0) then
  write(*,*) "Loaded velocity models"
  endif

  


sourceZ = 0
sourceX = 0
do rhs=1,NRHS
    sourceZ(rhs) = zi 
    sourceX(rhs) = xi + ds*(rhs-1)
    if (passive == 1) rayp(rhs) = sin(maxp*pi/180)/vel0(zi+nz*(xi-1))*(rhs-Np)
    if (myrank==0) write(*,*) rayp(rhs)
end do



!! set up global variables
!Lvl            = 16
slvl           = 0
rkmin          = 20                                       ! minimum rank to guarantee the numerical stability
tol            = 1.0E-4
Lpar           = nint(log(real(nprocs))/log(2.0)) + 1     ! Lpar = log_2(nprocs) + 1
mr2d           = 0
N_delta        = 50



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  1. 2D nested dissection preprocessing  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
t1 = MPI_Wtime()
 call NDpre_generate( myrank, tos, Lvl, Lpar, Nz, Nx, lengbTT, lenlocTT, gbTT, locTT, gbTlabel, locTlabel, gbNB, locNB, locHZ, locTZ, locHX, locTX, label, flag )
t2 = MPI_Wtime()
!if (myrank == 0) then
    !write(*,*) "2D nested dissection domain decomposition is done! wall time is ", t2-t1
!end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  2. BLACS contexts generating  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 call ctxt_generate( flag, lengbTT, gbTT, ctxt, nprow, npcol, myrow, mycol, mynodes ) 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!! find vel model for processor
    allocate(vel_pz(locTZ-locHZ+1,locTX-locHX+1))
    allocate(dvel_pz(locTZ-locHZ+1,locTX-locHX+1))
    allocate(rho(locTZ-locHZ+1+2*tos,locTX-locHX+1+2*tos))
    rho = 1.0E+0









do iw=w0,wf

!! Define initial conditions

Y = ZERO
Y0 = ZERO

if (passive == 1) then

taper=0.
do zz=101,Nx-100
  do xx=1,Nx
    taper(xx)=taper(xx)+exp(-real(xx-zz)**2/50**2.)
  enddo
enddo
taper = taper/maxval(taper,Nx)
do rhs = 1,NRHS
    do xx = 1,Nx
        do zz = 1,Nz
            tmp = real((zz-sourceZ(rhs))**2)
	    !tmp = real((zz-sourceZ(rhs))**2+(xx-sourceX(rhs))**2)
            if (tmp <= 25.0E+0) then
            Y(zz+Nz*(xx-1)+Nz*Nx*(rhs-1)) = taper(xx)*exp(RT*hx*xx*rayp(rhs)*real(freq(iw))*dt*10)*exp(-tmp/4.0)
	        Y0(zz+Nz*(xx-1)+Nz*Nx*(rhs-1)) = taper(xx)*exp(RT*hx*xx*rayp(rhs)*real(freq(iw))*dt*10)*exp(-tmp/4.0)	
            end if
        end do
    end do
end do


	 
 
 if (myrank == 0) then
  write(wnum,fmt='(I2)')iw
  sfile=trim(sfile0)//'s'//trim(wnum)
  
  dataw = ZERO
  do xx=1,Nx
   do rhs=1,NRHS
      dataw(xx + Nx*(rhs-1)) = Y(sourceZ(rhs) + Nz*(xx-1) + Nz*Nx*(rhs-1))
   enddo
  enddo
  
  call complexoutb(dataw,Nx*NRHS,sfile)
  
 endif


else
do rhs = 1,NRHS
    do xx = 1,Nx
        do zz = 1,Nz
            tmp = real((zz-sourceZ(rhs))**2+(xx-sourceX(rhs))**2)
            if (tmp <= 25.0E+0) then
           Y(zz+Nz*(xx-1)+Nz*Nx*(rhs-1)) = exp(-tmp/4.0)
	        Y0(zz+Nz*(xx-1)+Nz*Nx*(rhs-1)) = exp(-tmp/4.0)	
            end if
        end do
    end do
end do


endif


if (myrank==0) then
write(*,*) " "
write(*,*) "Propagating for frequency ",iw
write(*,*) "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%"
endif



    do xx = locHX,locTX
        do zz = locHZ,locTZ
	    vel_pz(zz-locHZ+1,xx-locHX+1) = vel(zz+Nz*(xx-1))  !starting model
        end do
    end do




if (myrank == 0) write(*,*) "Beginning to propagate"



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  3. generating the matrix A  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
t1 = MPI_Wtime()
if (JOB == 1) then        ! isotropic acoustic
    call A2D_acoustic( freq(iw), tos, Nz, Nx, N_delta, sigma0, hz, hx, locHZ, locTZ, locHX, locTX, label, vel_pz, rho, A )
endif
t2 = MPI_Wtime()

!if (myrank == 0) then
    !write(*,*) "local matrix A generation is done! Wall time is", t2-t1
!end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  4. factorization  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!t1 = MPI_Wtime()
 call factorization(A,slvl,Lpar,rkmin,tol,mr2d,flag,ctxt,nprow,npcol,myrow,mycol,mynodes, & 
                   lenlocTT,lengbTT,locTT,gbTT,locTlabel,gbTlabel,locNB,gbNB,locUL,locUR,locLL,locIPIV,gbUL,gbUR,gbLL,gbIPIV)
!t2 = MPI_Wtime()
if (myrank == 0) then
   !write(*,*) "The total factorization is done! Wall time is", t2-t1
end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  5. solution  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
lnx=locTX-locHX+1
lnz=locTZ-locHZ+1


! write(*,*) "debuga"
X = ZERO
do rhs = 1,NRHS
    do xx = locHX,locTX
        do zz = locHZ,locTZ
	    X(zz-locHz+1,xx-locHx+1,rhs) = Y(zz+Nz*(xx-1)+Nz*Nx*(rhs-1))  !!Y(zz,xx,rhs)	    
        end do
    end do
end do
if (locHZ > 1) then   ! CAUTION!!
    X(1:tos,:,:) = ZERO
end if
if (locHX > 1) then
    X(:,1:tos,:) = ZERO
end if

 !write(*,*) "debugb"
 
!t1 = MPI_Wtime()
 call mfsolve(X,NRHS,slvl,Lpar,flag,Nz,Nx,locHZ,locTZ,locHX,locTX,label,ctxt,nprow,npcol,myrow,mycol,mynodes, &
             lenlocTT,lengbTT,locTT,gbTT,locTlabel,gbTlabel,locNB,gbNB,locUL,locUR,locLL,locIPIV,gbUL,gbUR,gbLL,gbIPIV)
!t2 = MPI_Wtime()
if (myrank == 0) then
    !write(*,*) "The wall time of the solution to", NRHS, " RHS is ", t2-t1
end if















!! find vel model for processor

    do xx = locHX,locTX
        do zz = locHZ,locTZ
	    vel_pz(zz-locHZ+1,xx-locHX+1) = vel0(zz+Nz*(xx-1))  !starting model
        end do
    end do


if (myrank == 0) write(*,*) "Beginning to propagate"


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  3. generating the matrix A  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
t1 = MPI_Wtime()
if (JOB == 1) then        ! isotropic acoustic
    call A2D_acoustic( freq(iw), tos, Nz, Nx, N_delta, sigma0, hz, hx, locHZ, locTZ, locHX, locTX, label, vel_pz, rho, A )
endif
t2 = MPI_Wtime()

!if (myrank == 0) then
    !write(*,*) "local matrix A generation is done! Wall time is", t2-t1
!end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  4. factorization  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!t1 = MPI_Wtime()
 call factorization(A,slvl,Lpar,rkmin,tol,mr2d,flag,ctxt,nprow,npcol,myrow,mycol,mynodes, & 
                   lenlocTT,lengbTT,locTT,gbTT,locTlabel,gbTlabel,locNB,gbNB,locUL,locUR,locLL,locIPIV,gbUL,gbUR,gbLL,gbIPIV)
!t2 = MPI_Wtime()
if (myrank == 0) then
   !write(*,*) "The total factorization is done! Wall time is", t2-t1
end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  5. solution  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
lnx=locTX-locHX+1
lnz=locTZ-locHZ+1


! write(*,*) "debuga"
X = ZERO
do rhs = 1,NRHS
    do xx = locHX,locTX
        do zz = locHZ,locTZ
	    X0(zz-locHz+1,xx-locHx+1,rhs) = Y0(zz+Nz*(xx-1)+Nz*Nx*(rhs-1))  !!Y(zz,xx,rhs)
	    
        end do
    end do
end do
if (locHZ > 1) then   ! CAUTION!!
    X(1:tos,:,:) = ZERO
end if
if (locHX > 1) then
    X(:,1:tos,:) = ZERO
end if

 !write(*,*) "debugb"
 
!t1 = MPI_Wtime()
 call mfsolve(X0,NRHS,slvl,Lpar,flag,Nz,Nx,locHZ,locTZ,locHX,locTX,label,ctxt,nprow,npcol,myrow,mycol,mynodes, &
             lenlocTT,lengbTT,locTT,gbTT,locTlabel,gbTlabel,locNB,gbNB,locUL,locUR,locLL,locIPIV,gbUL,gbUR,gbLL,gbIPIV)
!t2 = MPI_Wtime()
if (myrank == 0) then
    !write(*,*) "The wall time of the solution to", NRHS, " RHS is ", t2-t1
end if







   headZ = locHZ
   if (headZ > 1) then
        headZ = headZ + tos
   end if
   headX = locHX
   if (headX > 1) then
       headX = headX + tos
   end if

  
do rhs = 1,NRHS
    do xx = headX,locTX
        do zz = headZ,locTZ
	    !!WRITE OUT BOTH I and dI TO SAME VARIABLE
	    YY(zz+Nz*(xx-1)+Nz*Nx*(rhs-1)) = X(zz-locHz+1,xx-locHx+1,rhs)-X0(zz-locHz+1,xx-locHx+1,rhs)
        end do
    end do
end do






 if (myrank == 0) then
  
  sfile=trim(sfile0)//trim(wnum)
  dataw = ZERO
  do xx=1,Nx
   do rhs=1,NRHS
      dataw(xx + Nx*(rhs-1)) = YY(sourceZ(rhs) + Nz*(xx-1) + Nz*Nx*(rhs-1)) - YY(sourceZ(rhs) + Nz*(xx-1) + Nz*Nx*(rhs-1))
   enddo
  enddo
  
  call complexoutb(dataw,Nx*NRHS,sfile)
  
 endif
 
enddo

deallocate(X,X0, Y,Y0,YY,vel,vel0,dv, freq, sourceX, sourceZ,dataw)


!! MPI Finalization
call MPI_Finalize(ierr)



end program modeling
