program migration

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!  2D massively parallel HSS multifontal Helmholtz solver: HelmPar2D  !!!!!!!!!!!!!!!
!!!!!!!!!!!           Shen Wang,  Jianlin Xia,  Maarten V. de Hoop              !!!!!!!!!!!!!!!
!!!!!!!!!!!                Purdue & ConocoPhillips, June 2010                   !!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

use NDpre2D
use context
use A2D 
use fact
use solution2D
use parser
!!use propagate_w

implicit none
include 'mpif.h'


!! declaring variables
integer                      ::  ierr, myrank, nprocs
integer                      ::  JOB, tos
integer                      ::  Lvl, slvl, Lpar, rkmin, mr2d
real(kind=4)                 ::  tol, dt, dw
real(kind=8)                 ::  t1, t2
integer                      ::  Nz, Nx, N_delta, NRHS, Nt, Nw,Nh
 complex(kind=4)              :: sigma0, hz, hx, tmpi,tmpdx,tmpdz
integer,allocatable          ::  sourceZ(:), sourceX(:),h(:)
real(kind=4),allocatable     ::  Istack(:),Is(:),I0(:),Ipr(:,:,:),vel(:),vel0(:), dv(:), ga(:), gz(:),gx(:), taper(:), N(:)
 complex(kind=4),allocatable  ::  Y(:),Y0(:), freq(:),Aw(:),dataw(:)
 complex(kind=4),allocatable  ::  X(:,:,:) !,Y2(:)
real(kind=4)                 ::  tmp, tmpz, tmpx,alpha
integer                      ::  rhs, zz, xx, pp, iw, w0, wf,xi,zi, ig,ds,dh,ihh,hh, iss, ix, iz, headZ, headX
 character(len=256)           ::  vfile,v0file,dvfile,sfile,sfile0,rfile,rfile0,ifile,ifile0,wnum 
 !!complex(kind=4),parameter   :: pi = (3.1415926E+0,0.0E+0)
 !!complex(kind=4),parameter   :: ONE = (1.0E+0,0.0E+0), ZERO = (0.0E+0,0.0E+0), RT = (0.0E+0,1.0E+0)
real(kind=4),allocatable     ::  vel_pz(:,:), vel_sz(:,:), rho(:,:)

!! MPI Initialization
 call MPI_Init(ierr)
 call MPI_Comm_rank( MPI_COMM_WORLD, myrank, ierr )
 call MPI_Comm_size( MPI_COMM_WORLD, nprocs, ierr )


!! set up variables
 call frompar('level',Lvl,10)
 call frompar('Nz',Nz,2048)
 call frompar('Nx',Nx,2048)
 !call frompar('Ns',NRHS,1)
 call frompar('Nt',Nt,2048)
 call frompar('Nw',Nw,55)
 call frompar('Nh',Nh,20)
 call frompar('dh',dh,2)
 call frompar('dx',tmp,250.); hx=tmp
 call frompar('dz',tmp,250.); hz=tmp
 call frompar('dt',dt,0.01)
 call frompar('ds',ds,10)
 call frompar('w0',w0,2)
 call frompar('wf',wf,Nw)
 call frompar('x0',xi,80)
 call frompar('z0',zi,110)
 call frompar('alpha',alpha,1.)
 call frompar('vfile',vfile)
 call frompar('v0file',v0file)
 call frompar('dvfile',dvfile)
 call frompar('sfile',sfile0)
 call frompar('rfile',rfile0,' ')
 call frompar('ifile',ifile0)


!! choosing jobs. 1: isotropic acoustic, 2: elliptic anisotropy, 3: VTI, 4. HTI, 5. TTI
JOB = 1
if (JOB == 5) then
    tos = 2
else
    tos = 1
end if

if (myrank==0) then
write(*,*) "Data source: ",trim(sfile0)
write(*,*) "Image file:  ",trim(ifile0)
endif


NRHS=int((Nx-2*xi)/ds + 1)
if (myrank==0) then
write(*,*) "Number of sources", NRHS
endif



allocate(Y(Nz*Nx*NRHS*2))
allocate(Is(Nz*Nx*NRHS))

!if (myrank==0) then
!allocate(Istack(Nz*Nx))
!allocate(Is(Nz*Nx*NRHS))
!allocate(I0(Nz*Nx*NRHS))
!allocate(N(Nz*Nx*NRHS))
!endif

allocate(dataw(Nx*NRHS))
allocate(freq(Nw),Aw(Nw))
allocate(sourceZ(NRHS),sourceX(NRHS))
allocate(vel(Nz*Nx))
allocate(vel0(Nz*Nx))
allocate(dv(Nz*Nx))
allocate(taper(Nx))
allocate(ga(8), gx(8), gz(8))
allocate(h(2*Nh-1))

dw = 2*PI/(Nt*dt)
do iw=1,Nw
  freq(iw)=(iw-1)*dw
  Aw(iw) = exp(-(real(wf+w0)/2 - real(iw))**2/(real(wf-w0)/5)**2)
enddo
sigma0 = 1.0*freq(wf)

sourceZ = 0
sourceX = 0
do rhs = 1,NRHS
    sourceZ(rhs) = zi 
    sourceX(rhs) = xi + ds*(rhs-1)
end do

do ihh=1,2*Nh-1
  h(ihh)=dh*(ihh-Nh)
enddo

NRHS = NRHS * 2 !! NOTE: TO ACCOMODATE SOURCE AND RECEIVER

!! read in velocity model

if (1 > 0) then
  call realinb(vel,Nx*Nz,vfile)
  call realinb(vel0,Nx*Nz,v0file)
  !call realinb(dv,Nx*Nz,dvfile)
  do xx=1,Nx*Nz
    vel(xx)=vel0(xx)+alpha*vel(xx)
  enddo
  if (myrank==0) then
  write(*,*) "Loaded velocity models"
  endif
else
    gx=(/350., 450., 550., 650., 350., 450., 550., 650./)
    gz=(/350., 350., 350., 350., 550., 550., 550., 500./)
    ga=(/1., 2., 2., 1., 2., 3., 3., 2./) 
    vel = 7200.0E+0
    !!dv = 0.0
    do xx=1,Nx
    	do zz=1,Nz
	do ig =1,8
	   tmp = 100*ga(ig)*real((zz-gz(ig))**2+(xx-gx(ig))**2)
	   vel(zz+Nz*(xx-1)) = vel(zz+Nz*(xx-1)) + exp(-tmp/2.)
	enddo
	enddo

    enddo
endif
if (myrank==0) then
write(*,*) "debug4",maxval(Istack,Nx*Nz)
endif

taper=0.
do zz=201,Nx-200
  do xx=1,Nx
    taper(xx)=taper(xx)+exp(-real(xx-zz)**2/50**2.)
  enddo
enddo
taper = taper/maxval(taper,Nx)

!Istack=0.
Is=0.
!I0=0.
!Ih=0.



!! set up global variables
!Lvl            = 16
slvl           = 0
rkmin          = 20                                       ! minimum rank to guarantee the numerical stability
tol            = 1.0E-4
Lpar           = nint(log(real(nprocs))/log(2.0)) + 1     ! Lpar = log_2(nprocs) + 1
mr2d           = 0
N_delta        = 50


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  1. 2D nested dissection preprocessing  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
t1 = MPI_Wtime()
 call NDpre_generate( myrank, tos, Lvl, Lpar, Nz, Nx, lengbTT, lenlocTT, gbTT, locTT, gbTlabel, locTlabel, gbNB, locNB, locHZ, locTZ, locHX, locTX, label, flag )
t2 = MPI_Wtime()
if (myrank == 0) then
    write(*,*) "2D nested dissection domain decomposition is done! wall time is ", t2-t1
end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  2. BLACS contexts generating  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 call ctxt_generate( flag, lengbTT, gbTT, ctxt, nprow, npcol, myrow, mycol, mynodes ) 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!! find vel model for processor
  allocate(vel_pz(locTZ-locHZ+1,locTX-locHX+1))
    do xx = locHX,locTX
        do zz = locHZ,locTZ
	    vel_pz(zz-locHZ+1,xx-locHX+1) = vel(zz+Nz*(xx-1))  !vel(zz,xx)	    
        end do
    end do
  
  allocate(rho(locTZ-locHZ+1+2*tos,locTX-locHX+1+2*tos))
  rho = 1.0E+0


  allocate(Ipr(locTZ-locHZ+1,locTX-locHX+1,NRHS))
  Ipr=0.


do iw=w0,wf

!! Define initial conditions

Y = ZERO
!!Y0 = ZERO

if (myrank==0) then
write(*,*) " "
write(*,*) "Propagating for frequency ",iw
write(*,*) "********************************"
endif


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  3. generating the matrix A  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
t1 = MPI_Wtime()
if (JOB == 1) then        ! isotropic acoustic
    call A2D_acoustic( freq(iw), tos, Nz, Nx, N_delta, sigma0, hz, hx, locHZ, locTZ, locHX, locTX, label, vel_pz, rho, A )
endif
t2 = MPI_Wtime()

if (myrank == 0) then
    write(*,*) "local matrix A generation is done! Wall time is", t2-t1
end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  4. factorization  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
t1 = MPI_Wtime()
 call factorization(A,slvl,Lpar,rkmin,tol,mr2d,flag,ctxt,nprow,npcol,myrow,mycol,mynodes, & 
                   lenlocTT,lengbTT,locTT,gbTT,locTlabel,gbTlabel,locNB,gbNB,locUL,locUR,locLL,locIPIV,gbUL,gbUR,gbLL,gbIPIV)
t2 = MPI_Wtime()
if (myrank == 0) then
    write(*,*) "The total factorization is done! Wall time is", t2-t1
end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  4A. define initial wavefields !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  write(wnum,fmt='(I2)')iw
  sfile=trim(sfile0)//trim(wnum)
  call complexinb(dataw,Nx*NRHS/2,sfile)

  do rhs=1,NRHS/2
    do xx=1,Nx
      do zz=2,Nz-1
         tmpz = real((zz-sourceZ(rhs))**2)
	 tmpx = real((xx-sourceX(rhs))**2)
	 if (tmpz<=16.0) then
            Y(sourceZ(rhs)+1+Nz*(xx-1)+Nz*Nx*(rhs-1+NRHS/2))=Aw(iw)*taper(xx)*exp(-tmpz/16.0)*conjg(dataw((xx-1)+Nx*(rhs-1)))
            Y(sourceZ(rhs)-1+Nz*(xx-1)+Nz*Nx*(rhs-1+NRHS/2))=-Aw(iw)*taper(xx)*exp(-tmpz/16.0)*conjg(dataw((xx-1)+Nx*(rhs-1)))	    
            !!Y(sourceZ(rhs)+Nz*(xx-1)+Nz*Nx*(rhs-1+NRHS/2))=Aw(iw)*taper(xx)*exp(-tmpz/16.0)*conjg(dataw((xx-1)+Nx*(rhs-1)))
	 endif
         if (tmpx+tmpz <= 25.0E+0) then
	    Y(zz+Nz*(xx-1)+Nz*Nx*(rhs-1)) = Aw(iw)*exp(-(tmpx+tmpz)/4.0)	    
         end if
      enddo
    enddo
  enddo
  if (myrank==0) then
  !write(*,*) "debug6"
  endif
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  5. solution  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
allocate(X(locTZ-locHZ+1,locTX-locHX+1,NRHS))
X = ZERO
do rhs = 1,NRHS
    do xx = locHX,locTX
        do zz = locHZ,locTZ
	    X(zz-locHz+1,xx-locHx+1,rhs) = Y(zz+Nz*(xx-1)+Nz*Nx*(rhs-1))  !!Y(zz,xx,rhs)
        end do
    end do
end do
if (locHZ > 1) then   ! CAUTION!!
    X(1:tos,:,:) = ZERO
end if
if (locHX > 1) then
    X(:,1:tos,:) = ZERO
end if

t1 = MPI_Wtime()
 call mfsolve(X,NRHS,slvl,Lpar,flag,Nz,Nx,locHZ,locTZ,locHX,locTX,label,ctxt,nprow,npcol,myrow,mycol,mynodes, &
             lenlocTT,lengbTT,locTT,gbTT,locTlabel,gbTlabel,locNB,gbNB,locUL,locUR,locLL,locIPIV,gbUL,gbUR,gbLL,gbIPIV)
t2 = MPI_Wtime()
if (myrank == 0) then
    write(*,*) "The wall time of the solution to", NRHS, " RHS is ", t2-t1
end if



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


 !if (myrank == 0) then

t1 = MPI_Wtime()
  do xx=1,locTX-locHX+1
   do zz = 1,locTZ-locHZ+1
    do rhs=1,NRHS/2
      
      tmpi = X(zz,xx,rhs) * X(zz,xx,rhs+NRHS/2)
      !tmpdx = (Y(zz+1*Nz+xx+rhs)-Y(zz-1*Nz+xx+rhs))*(Y(zz+1*Nz+xx+rhs+Nz*Nx*NRHS/2)-Y(zz-1*Nz+xx+rhs+Nz*Nx*NRHS/2))/(4*hx**2) 
      !tmpdz = (Y(zz+1+xx+rhs)-Y(zz-1+xx+rhs))*(Y(zz+1+xx+rhs+Nz*Nx*NRHS/2)-Y(zz-1+xx+rhs+Nz*Nx*NRHS/2))/(4*hz**2)
      
      !Istack(zz+xx) = Istack(zz+xx) + real(PI**(-1)*2*RT*freq(iw)**(-3)*(freq(iw)**2*vel(zz+xx)**(-2)*tmpi - tmpdx - tmpdz))
      !Is(zz+xx+rhs) = Is(zz+xx+rhs) + real(PI**(-1)*2*RT*freq(iw)**(-3)*(freq(iw)**2*vel(zz+xx)**(-2)*tmpi - tmpdx - tmpdz))
      Ipr(zz,xx,rhs) = Ipr(zz,xx,rhs) + real(tmpi)
      !N(zz+xx+rhs) = N(zz+xx+rhs) + real(Y(zz+xx+rhs)*conjg(Y(zz+xx+rhs)))

    enddo
   enddo
  enddo
  t2 = MPI_Wtime()
  if (myrank == 0) then
    write(*,*) "The wall time for imaging is ", t2-t1
  endif
 !endif
 
 call factors_dealloc( lenlocTT, locUL, locUR, locLL, locIPIV, Lpar-1, gbUL, gbUR, gbLL, gbIPIV )
 deallocate(X,locUL, locUR, locLL, locIPIV, gbUL, gbUR, gbLL, gbIPIV )  


enddo !!!%%%%%%%%%%%%%%%%%%%%%%%END LOOP OVER FREQUENCY%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!!!

   headZ = locHZ
   if (headZ > 1) then
        headZ = headZ + tos
   end if
   headX = locHX
   if (headX > 1) then
       headX = headX + tos
   end if
  if (myrank == 0) then
    write(*,*) "debug1"
  endif  
  allocate(I0(Nz*Nx*NRHS/2))
do rhs = 1,NRHS/2
    do xx = headX,locTX
        do zz = headZ,locTZ
	    !!X(zz-locHz+1,xx-locHx+1,rhs) = Y(zz,xx,rhs)
	    I0(zz+Nz*(xx-1)+Nz*Nx*(rhs-1)) = Ipr(zz-locHz+1,xx-locHx+1,rhs)
        end do
    end do
end do
  if (myrank == 0) then
    write(*,*) "debug2"
  endif
  
  
t1 = MPI_Wtime()
!!write(*,*) myrank,locHZ, locTZ,locHX,locTX
 call MPI_Barrier(MPI_COMM_WORLD,ierr)
   if (myrank == 0) then
    write(*,*) "debug3"
  endif
 Is=0.
 call MPI_REDUCE(I0,Is,Nx*Nz*NRHS/2,MPI_REAL,MPI_SUM,0,MPI_COMM_WORLD,ierr)

 deallocate(I0)
 
t2 = MPI_Wtime()
 if (myrank == 0) then
    write(*,*) "The wall time of the MPI_reduce", t2-t1
 endif

  
  if (myrank==0) then
    ifile = trim(ifile0)//'PR'
    !call realoutb(Istack,Nz*Nx,ifile0)
    call realoutb(Is,Nz*Nx*NRHS/2,ifile)
    !ifile = trim(ifile0)//'0'
    !call realoutb(I0,Nz*Nx*NRHS/2,ifile)
    !!call realoutb(N,Nz*Nx*NRHS/2,rfile0)
    !!ifile=trim(ifile0)//'H'
    !!call realoutb(Ih,Nz*Nx*(2*Nh-1)) 
 endif


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  6. deallocating factors  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 deallocate(gbTT,locTT,gbTlabel,locTlabel,label,locNB,gbNB)
 deallocate(ctxt, nprow, npcol, myrow, mycol, mynodes)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !if (myrank==0) then
  !  deallocate(Istack,Is,I0,N)
  !endif
deallocate(Y,Is,vel,vel0,dv, freq, Aw, sourceX, sourceZ,h,dataw, taper)



!! MPI Finalization
 call MPI_Finalize(ierr)



end program migration
