module sol_kernel

use partition
use context
use fact
implicit none
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  multifrontal solution kernel subroutine  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


    !! local RHS
    complex(kind=4),allocatable,private :: locB(:,:)

    !! redistribution RHS
    type(redist_struct),allocatable,private :: gbBT(:)


contains


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  multifrontal solver  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine mfsolve_kernel(locUL,locUR,locLL,locIPIV,gbUL,gbUR,gbLL,gbIPIV,slvl,Lpar,flag,ctxt,nprow,npcol,myrow,mycol,mynodes,lenlocTT,lengbTT,locTT,gbTT,locTlabel,gbTlabel,locNB,gbNB,NRHS,locB,gbBT)

        type(UL_struct)                 :: locUL(lenlocTT), gbUL(lengbTT)
        type(UR_struct)                 :: locUR(lenlocTT), locLL(lenlocTT), gbUR(lengbTT), gbLL(lengbTT)
        type(IPIV_struct)               :: locIPIV(lenlocTT), gbIPIV(lengbTT)
        integer                         :: slvl, Lpar, flag, lenlocTT, lengbTT, NRHS
        integer                         :: ctxt(Lpar), nprow(Lpar), npcol(Lpar), myrow(Lpar), mycol(Lpar), mynodes(Lpar)
        type(TT_struct)                 :: locTT(lenlocTT), gbTT(lengbTT)
        integer                         :: locTlabel(lenlocTT+1), gbTlabel(lengbTT+1)
        type(NB_struct)                 :: locNB(lenlocTT), gbNB(lengbTT)
        complex(kind=4),allocatable     :: locB(:,:)
        type(redist_struct)             :: gbBT(lengbTT)

        integer      :: stat = 0
        real(kind=8) :: flops = 0.0D+0

        !! 1. local solution: Ly = b
        call loc_solL( slvl, Lpar, flag, lenlocTT, locTT, locTlabel, locNB, locUL, locLL, locIPIV, locB, gbBT, NRHS, stat, flops )


        !! 2. global solution: Ly = b, Ux = y
        call gb_sol( slvl, flag, lengbTT, gbTT, gbTlabel, gbNB, ctxt, nprow, npcol, myrow, mycol, mynodes, gbUL, gbUR, gbLL, gbIPIV, gbBT, NRHS, stat, flops )


        !! 3. local solution: Ux = y
        call loc_solU( slvl, Lpar, flag, lenlocTT, locTT, locTlabel, locNB, locUL, locUR, locB, gbBT, NRHS, stat, flops )
        

        !! 4. statistics
        if (stat /= 0) then
            call dgsum2d( ctxt(1), 'A', 'I', 1, 1, flops, 1, 0, 0 )
            if (flag == 1) then
                write(*,*) "solution flops is: ", flops
            end if
        end if


        return
    end subroutine mfsolve_kernel


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  1. local solution: Ly = b  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine loc_solL( slvl, Lpar, flag, lenlocTT, locTT, locTlabel, locNB, locUL, locLL, locIPIV, locB, gbBT, NRHS, stat, flops )

        integer                         :: slvl, Lpar, flag, lenlocTT, NRHS
        type(TT_struct)                 :: locTT(lenlocTT)
        integer                         :: locTlabel(lenlocTT+1)
        type(NB_struct),target          :: locNB(lenlocTT)
        type(UL_struct)                 :: locUL(lenlocTT)
        type(UR_struct)                 :: locLL(lenlocTT)
        type(IPIV_struct)               :: locIPIV(lenlocTT)
        complex(kind=4),allocatable     :: locB(:,:)
        type(redist_struct)             :: gbBT(*)
        integer                         :: stat
        real(kind=8)                    :: flops

        integer                     :: level, Ntmp, Nhead, Ntail, Ptmp, Phead, Ptail, szUL, szLR, rkLL, counter, info
        type(NB_struct),pointer     :: tail
        complex(kind=4),allocatable :: tmp(:,:)

        do Ntmp = 1,lenlocTT

            level = locTT(Ntmp)%LL
            Nhead = locTlabel(Ntmp)
            Ntail = locTlabel(Ntmp+1)-1
            szUL  = Ntail - Nhead + 1

            ! triangular solver for Nhead:Ntail
            if (level <= slvl-Lpar+1) then
                call loctrs_compr( -1, 0, NRHS, locUL(Ntmp), locIPIV(Ntmp), locB(Nhead:Ntail,1:NRHS), szUL )
            else
                call claswp( NRHS, locB(Nhead:Ntail,:), szUL, 1, szUL, locIPIV(Ntmp)%ary1, 1 )
                call ctrtrs( 'L', 'N', 'U', szUL, NRHS, locUL(Ntmp)%ary11, szUL, locB(Nhead:Ntail,:), szUL, info )
            end if
            if (stat /= 0) then
                flops = flops + dble(szUL)**2.0 * dble(NRHS)
            end if

            if (locNB(Ntmp)%Ohead /= 0) then

                ! allocating tmp if LL is compressed
                if (level <= slvl-Lpar+1) then
                    rkLL = locLL(Ntmp)%rk
                    allocate(tmp(rkLL,NRHS))
                    tmp = ZERO
                    call cgemm( 'T', 'N', rkLL, NRHS, szUL, ONE, locLL(Ntmp)%Q1, szUL, locB(Nhead:Ntail,1:NRHS), szUL, ONE, tmp, rkLL )
                    if (stat /= 0) then
                        flops = flops + dble(rkLL) * dble(NRHS) * 2.0*dble(szUL)
                    end if
                end if

                counter = 1
                tail => locNB(Ntmp)
                do while (associated(tail))
                    Ptmp = tail%Node
                    szLR = tail%Otail - tail%Ohead + 1
                    if (Ptmp /= 0) then
                        Phead = tail%Ohead
                        Ptail = tail%Otail
                        if (level <= slvl-Lpar+1) then
                            call cgemm( 'T', 'N', szLR, NRHS, rkLL, -ONE,   locLL(Ntmp)%R(:,counter:counter+szLR-1), rkLL,                 tmp, rkLL, ONE, locB(Phead:Ptail,:), szLR )
                        else
                            call cgemm( 'T', 'N', szLR, NRHS, szUL, -ONE, locLL(Ntmp)%ary(:,counter:counter+szLR-1), szUL, locB(Nhead:Ntail,:), szUL, ONE, locB(Phead:Ptail,:), szLR )
                        end if
                    else
                        Phead = tail%Ohead - locTlabel(lenlocTT+1) + 1
                        Ptail = tail%Otail - locTlabel(lenlocTT+1) + 1
                        if (level <= slvl-Lpar+1) then
                            call cgemm( 'T', 'N', szLR, NRHS, rkLL, -ONE,   locLL(Ntmp)%R(:,counter:counter+szLR-1), rkLL,                 tmp, rkLL, ONE, gbBT(flag)%ary1(Phead:Ptail,:), szLR )
                        else
                            call cgemm( 'T', 'N', szLR, NRHS, szUL, -ONE, locLL(Ntmp)%ary(:,counter:counter+szLR-1), szUL, locB(Nhead:Ntail,:), szUL, ONE, gbBT(flag)%ary1(Phead:Ptail,:), szLR )
                        end if
                    end if
                    counter = counter + szLR
                    tail => tail%next
                end do
                if (stat /= 0) then
                    if (level <= slvl-Lpar+1) then
                        flops = flops + dble(szLR) * dble(NRHS) * 2.0*dble(rkLL)
                    else
                        flops = flops + dble(szLR) * dble(NRHS) * 2.0*dble(szUL)
                    end if
                end if

                if (level <= slvl-Lpar+1) then
                    deallocate(tmp)
                end if

            end if

        end do

        return
    end subroutine loc_solL



    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  2. global solution: Ly = b and Ux = y  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine gb_sol( slvl, flag, lengbTT, gbTT, gbTlabel, gbNB, ctxt, nprow, npcol, myrow, mycol, mynodes, gbUL, gbUR, gbLL, gbIPIV, gbBT, NRHS, stat, flops )

        integer                           :: slvl, flag, lengbTT, NRHS
        type(TT_struct)                   :: gbTT(lengbTT)
        integer                           :: gbTlabel(lengbTT+1)
        type(NB_struct),target            :: gbNB(lengbTT)
        integer                           :: ctxt(*), nprow(*), npcol(*), myrow(*), mycol(*), mynodes(*)
        type(UL_struct)                   :: gbUL(lengbTT)
        type(UR_struct)                   :: gbUR(lengbTT), gbLL(lengbTT)
        type(IPIV_struct)                 :: gbIPIV(lengbTT)
        type(redist_struct)               :: gbBT(lengbTT)
        integer                           :: stat
        real(kind=8)                      :: flops

        type(UR_struct),allocatable     :: gbB(:)
        integer                         :: Lvl, level, Ntmp, kid, szUL, szLR, sz1, sz2, sz, rk1, rk2, rkUR, rkLL, LOCR, LOCC, kk, info
        integer,external                :: numroc
        integer                         :: descIPIV(9,2), descUL(9,6), descUR(9,3), descLL(9,3), descB(9,2), descBT(9), desc1(9), desc2(9), descTMP(9)
        integer                         :: IWORK(1)
        type(NB_struct),pointer         :: tail
        complex(kind=4),allocatable     :: tmp(:,:) 
        
        
        Lvl = gbTT(1)%LL - 1
        allocate(gbB(Lvl))

        !***********************************************  Ly = b  *********************************************!
        do level = Lvl,1,-1

            Ntmp = mynodes(level)
            call blacs_barrier(ctxt(level),'a')

            !! 1. allocating the memory for the gbB and gbBT
            szUL = gbTlabel(Ntmp+1) - gbTlabel(Ntmp)
            LOCC = numroc( NRHS, NB, mycol(level), CSRC, npcol(level) )

            if (level <= slvl) then

                sz1 = szUL / 2
                sz2 = szUL - sz1
                rk1 = gbUL(level)%rk1
                rk2 = gbUL(level)%rk2

                LOCR = numroc( sz1, MB, myrow(level), RSRC, nprow(level) )
                call descset( descUL(:,1), sz1, sz1, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )
                call descset( descUL(:,2), sz1, rk1, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )
                call descset( descUL(:,4), sz1, rk2, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )
                call descset( descIPIV(:,1), sz1+MB*nprow(level), 1, MB, 1, RSRC, mycol(level), ctxt(level), MB+LOCR )
                allocate( gbB(level)%Q1(LOCR,LOCC) )
                gbB(level)%Q1 = ZERO
                call descset( descB(:,1), sz1, NRHS, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )

                LOCR = numroc( rk1, MB, myrow(level), RSRC, nprow(level) )
                call descset( descUL(:,3), rk1, sz2, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )

                LOCR = numroc( rk2, MB, myrow(level), RSRC, nprow(level) )
                call descset( descUL(:,5), rk2, sz2, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )

                LOCR = numroc( sz2, MB, myrow(level), RSRC, nprow(level) )
                call descset( descUL(:,6), sz2, sz2, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )
                call descset( descIPIV(:,2), sz2+MB*nprow(level), 1, MB, 1, RSRC, mycol(level), ctxt(level), MB+LOCR )
                allocate( gbB(level)%Q2(LOCR,LOCC) )
                gbB(level)%Q2 = ZERO
                call descset( descB(:,2), sz2, NRHS, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )

            else

                LOCR = numroc( szUL, MB, myrow(level), RSRC, nprow(level) )
                call descset( descUL(:,1), szUL, szUL, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )
                call descset( descIPIV(:,1), szUL+MB*nprow(level), 1, MB, 1, RSRC, mycol(level), ctxt(level), MB+LOCR )
                allocate( gbB(level)%ary(LOCR,LOCC) )
                gbB(level)%ary = ZERO
                call descset( descB(:,1), szUL, NRHS, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )

            end if

            if (Ntmp /= lengbTT) then

                szLR = 0
                tail => gbNB(Ntmp)
                do while (associated(tail))
                    szLR = szLR + tail%Otail - tail%Ohead + 1
                    tail => tail%next
                end do

                if (level <= slvl) then

                    rkUR = gbUR(level)%rk
                    rkLL = gbLL(level)%rk

                    LOCR = numroc( sz1, MB, myrow(level), RSRC, nprow(level) )
                    call descset( descUR(:,1), sz1, rkUR, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )
                    call descset( descLL(:,1), sz1, rkLL, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )

                    LOCR = numroc( sz2, MB, myrow(level), RSRC, nprow(level) )
                    call descset( descUR(:,2), sz2, rkUR, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )
                    call descset( descLL(:,2), sz2, rkLL, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )

                    LOCR = numroc( rkUR, MB, myrow(level), RSRC, nprow(level) )
                    call descset( descUR(:,3), rkUR, szLR, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )

                    LOCR = numroc( rkLL, MB, myrow(level), RSRC, nprow(level) )
                    call descset( descLL(:,3), rkLL, szLR, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )

                else

                    LOCR = numroc( szUL, MB, myrow(level), RSRC, nprow(level) )
                    call descset( descUR(:,1), szUL, szLR, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )
                    descLL(:,1) = descUR(:,1)

                end if

                LOCR = numroc( szLR, MB, myrow(level), RSRC, nprow(level) )
                allocate( gbBT(Ntmp)%ary1(LOCR,LOCC) )
                gbBT(Ntmp)%ary1 = ZERO
                call descset( descBT, szLR, NRHS, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )

            end if

            !! 2. redistribution of gbBT(kid) to gbBT(Ntmp)
            do kk = 1,2

                kid = gbTT(Ntmp)%kids(kk)

                sz = 0
                tail => gbNB(kid)
                do while (associated(tail))
                    sz = sz + tail%Otail - tail%Ohead + 1
                    tail => tail%next
                end do

                ! desc1
                if (kk == 1) then
                    if (flag <= gbTT(Ntmp)%kids(1)) then
                        LOCR = numroc( sz, MB, myrow(level+1), RSRC, nprow(level+1) )
                        call descset( desc1, sz, NRHS, MB, NB, RSRC, CSRC, ctxt(level+1), max(1,LOCR) )
                    else
                        call descset( desc1, sz, NRHS, MB, NB, RSRC, CSRC, -1, 1 )
                    end if
                else
                    if (flag <= gbTT(Ntmp)%kids(1)) then
                        call descset( desc1, sz, NRHS, MB, NB, RSRC, CSRC, -1, 1 )
                    else
                        LOCR = numroc( sz, MB, myrow(level+1), RSRC, nprow(level+1) )
                        call descset( desc1, sz, NRHS, MB, NB, RSRC, CSRC, ctxt(level+1), max(1,LOCR) )
                    end if
                end if

                ! desc2
                LOCR = numroc(   sz, MB, myrow(level), RSRC, nprow(level) )
                call descset( desc2, sz, NRHS, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )
                allocate(gbBT(kid)%ary2(LOCR,LOCC))
                gbBT(kid)%ary2 = ZERO

                ! redistribution
                call pcgemr2d( sz, NRHS, gbBT(kid)%ary1, 1, 1, desc1, gbBT(kid)%ary2, 1, 1, desc2, ctxt(level) )
                if (allocated(gbBT(kid)%ary1)) then
                    deallocate(gbBT(kid)%ary1)
                end if

                ! forward extend-add for the RHS
                if (level <= slvl) then
                    call pextadd_compr( 1, NRHS, gbB(level)%Q1, descB(:,1), gbB(level)%Q2, descB(:,2), gbBT(Ntmp)%ary1, descBT, gbBT(kid)%ary2, desc2, Ntmp, kid, gbNB )
                else
                    call pextadd_RHS( 1, NRHS, gbB(level)%ary, descB(:,1), gbBT(Ntmp)%ary1, descBT, gbBT(kid)%ary2, desc2, Ntmp, kid, gbNB )
                end if
                if (allocated(gbBT(kid)%ary2)) then
                    deallocate(gbBT(kid)%ary2)
                end if

                if (stat /= 0) then
                    if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                        flops = flops + dble(sz) * dble(NRHS) * 2.0
                    end if
                end if

            end do

            !! 3. triangular solver
            if (level <= slvl) then
                call gbtrs_compr(-1,0,NRHS,ctxt(level),nprow(level),npcol(level),myrow(level),mycol(level),gbUL(level),descUL,gbIPIV(level),descIPIV,gbB(level)%Q1,descB(:,1),gbB(level)%Q2,descB(:,2))
            else
                call pclapiv( 'F', 'R', 'C', szUL, NRHS, gbB(level)%ary, 1, 1, descB(:,1), gbIPIV(level)%ary1, 1, 1, descIPIV(:,1), IWORK )
                call pctrtrs( 'L', 'N', 'U', szUL, NRHS, gbUL(level)%ary11, 1, 1, descUL(:,1), gbB(level)%ary, 1, 1, descB(:,1), info )
            end if
            if (stat /= 0) then
                if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                    flops = flops + dble(szUL)**2.0 * dble(NRHS)
                end if
            end if

            if (Ntmp /= lengbTT) then
                if (level <= slvl) then
                    LOCR = numroc( rkLL, MB, myrow(level), RSRC, nprow(level) )
                    allocate(tmp(LOCR,LOCC))
                    tmp = ZERO
                    call descset( descTMP, rkLL, NRHS, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )
                    call pcgemm( 'T', 'N', rkLL, NRHS,  sz1,  ONE, gbLL(level)%Q1, 1, 1, descLL(:,1), gbB(level)%Q1, 1, 1, descB(:,1), ONE, tmp, 1, 1, descTMP )
                    call pcgemm( 'T', 'N', rkLL, NRHS,  sz2,  ONE, gbLL(level)%Q2, 1, 1, descLL(:,2), gbB(level)%Q2, 1, 1, descB(:,2), ONE, tmp, 1, 1, descTMP )
                    call pcgemm( 'T', 'N', szLR, NRHS, rkLL, -ONE, gbLL(level)%R,  1, 1, descLL(:,3), tmp, 1, 1, descTMP, ONE, gbBT(Ntmp)%ary1, 1, 1, descBT )
                    deallocate(tmp)
                    if (stat /= 0) then
                        if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                            flops = flops + dble(rkLL) * dble(NRHS) * 2.0*dble(szUL) + dble(szLR) * dble(NRHS) * 2.0*dble(rkLL)
                        end if
                    end if
                else
                    call pcgemm( 'T', 'N', szLR, NRHS, szUL, -ONE, gbLL(level)%ary, 1, 1, descLL(:,1), gbB(level)%ary, 1, 1, descB(:,1), ONE, gbBT(Ntmp)%ary1, 1, 1, descBT )
                    if (stat /= 0) then
                        if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                            flops = flops + dble(szLR) * dble(NRHS) * 2.0*dble(szUL)
                        end if
                    end if
                end if
            end if

        end do


        !********************************************  Ux = y  ***************************************************!
        do level = 1,Lvl

            Ntmp = mynodes(level)
            call blacs_barrier(ctxt(level),'a')

            !! 1. allocating the memory for the gbB and gbBT
            szUL = gbTlabel(Ntmp+1) - gbTlabel(Ntmp)
            LOCC = numroc( NRHS, NB, mycol(level), CSRC, npcol(level) )

            if (level <= slvl) then

                sz1 = szUL / 2
                sz2 = szUL - sz1
                rk1 = gbUL(level)%rk1
                rk2 = gbUL(level)%rk2

                LOCR = numroc( sz1, MB, myrow(level), RSRC, nprow(level) )
                call descset( descUL(:,1), sz1, sz1, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )
                call descset( descUL(:,2), sz1, rk1, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )
                call descset( descUL(:,4), sz1, rk2, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )
                call descset( descB(:,1), sz1, NRHS, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )

                LOCR = numroc( rk1, MB, myrow(level), RSRC, nprow(level) )
                call descset( descUL(:,3), rk1, sz2, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )

                LOCR = numroc( rk2, MB, myrow(level), RSRC, nprow(level) )
                call descset( descUL(:,5), rk2, sz2, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )

                LOCR = numroc( sz2, MB, myrow(level), RSRC, nprow(level) )
                call descset( descUL(:,6), sz2, sz2, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )                
                call descset( descB(:,2), sz2, NRHS, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )

            else

                LOCR = numroc( szUL, MB, myrow(level), RSRC, nprow(level) )
                call descset( descUL(:,1), szUL, szUL, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )
                call descset(  descB(:,1), szUL, NRHS, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )

            end if

            if (Ntmp /= lengbTT) then

                szLR = 0
                tail => gbNB(Ntmp)
                do while (associated(tail))
                    szLR = szLR + tail%Otail - tail%Ohead + 1
                    tail => tail%next
                end do

                if (level <= slvl) then

                    rkUR = gbUR(level)%rk
                    rkLL = gbLL(level)%rk

                    LOCR = numroc( sz1, MB, myrow(level), RSRC, nprow(level) )
                    call descset( descUR(:,1), sz1, rkUR, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )
                    call descset( descLL(:,1), sz1, rkLL, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )

                    LOCR = numroc( sz2, MB, myrow(level), RSRC, nprow(level) )
                    call descset( descUR(:,2), sz2, rkUR, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )
                    call descset( descLL(:,2), sz2, rkLL, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )

                    LOCR = numroc( rkUR, MB, myrow(level), RSRC, nprow(level) )
                    call descset( descUR(:,3), rkUR, szLR, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )

                    LOCR = numroc( rkLL, MB, myrow(level), RSRC, nprow(level) )
                    call descset( descLL(:,3), rkLL, szLR, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )

                else

                    LOCR = numroc( szUL, MB, myrow(level), RSRC, nprow(level) )
                    call descset( descUR(:,1), szUL, szLR, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )
                    descLL(:,1) = descUR(:,1)

                end if

                LOCR = numroc( szLR, MB, myrow(level), RSRC, nprow(level) )
                call descset( descBT, szLR, NRHS, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )

            end if

            !! 2. triangular solver
            if (Ntmp /= lengbTT) then
                if (level <= slvl) then
                    LOCR = numroc( rkUR, MB, myrow(level), RSRC, nprow(level) )
                    allocate(tmp(LOCR,LOCC))
                    tmp = ZERO
                    call descset( descTMP, rkUR, NRHS, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )
                    call pcgemm( 'N', 'N', rkUR, NRHS, szLR,  ONE,  gbUR(level)%R, 1, 1, descUR(:,3), gbBT(Ntmp)%ary1, 1, 1, descBT, ONE, tmp, 1, 1, descTMP )
                    call pcgemm( 'N', 'N',  sz1, NRHS, rkUR, -ONE, gbUR(level)%Q1, 1, 1, descUR(:,1), tmp, 1, 1, descTMP, ONE, gbB(level)%Q1, 1, 1, descB(:,1) )
                    call pcgemm( 'N', 'N',  sz2, NRHS, rkUR, -ONE, gbUR(level)%Q2, 1, 1, descUR(:,2), tmp, 1, 1, descTMP, ONE, gbB(level)%Q2, 1, 1, descB(:,2) )
                    deallocate(tmp)
                    if (stat /= 0) then
                        if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                            flops = flops + dble(rkUR) * dble(NRHS) * 2.0*dble(szLR) + dble(szUL) * dble(NRHS) * 2.0*dble(rkUR)
                        end if
                    end if
                else
                    call pcgemm( 'N', 'N', szUL, NRHS, szLR, -ONE, gbUR(level)%ary, 1, 1, descUR(:,1), gbBT(Ntmp)%ary1, 1, 1, descBT, ONE, gbB(level)%ary, 1, 1, descB(:,1) )
                    if (stat /= 0) then
                        if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                            flops = flops + dble(szUL) * dble(NRHS) * 2.0*dble(szLR)
                        end if
                    end if
                end if
            end if

            if (level <= slvl) then
                call gbtrs_compr(1,0,NRHS,ctxt(level),nprow(level),npcol(level),myrow(level),mycol(level),gbUL(level),descUL,gbIPIV(level),descIPIV,gbB(level)%Q1,descB(:,1),gbB(level)%Q2,descB(:,2))
            else
                call pctrtrs( 'U', 'N', 'N', szUL, NRHS, gbUL(level)%ary11, 1, 1, descUL(:,1), gbB(level)%ary, 1, 1, descB(:,1), info )
            end if
            if (stat /= 0) then
                if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                    flops = flops + dble(szUL)**2.0 * dble(NRHS)
                end if
            end if

            !! 3. redistribution to gbBT(kid)%ary2
            do kk = 1,2

                kid = gbTT(Ntmp)%kids(kk)

                sz = 0
                tail => gbNB(kid)
                do while (associated(tail))
                    sz = sz + tail%Otail - tail%Ohead + 1
                    tail => tail%next
                end do

                ! desc1
                if (kk == 1) then
                    if (flag <= gbTT(Ntmp)%kids(1)) then
                        LOCR = numroc(   sz, MB, myrow(level+1), RSRC, nprow(level+1) )
                        LOCC = numroc( NRHS, NB, mycol(level+1), CSRC, npcol(level+1) )
                        allocate(gbBT(kid)%ary1(LOCR,LOCC))
                        gbBT(kid)%ary1 = ZERO
                        call descset( desc1, sz, NRHS, MB, NB, RSRC, CSRC, ctxt(level+1), max(1,LOCR) )
                    else
                        call descset( desc1, sz, NRHS, MB, NB, RSRC, CSRC, -1, 1 )
                    end if
                else
                    if (flag <= gbTT(Ntmp)%kids(1)) then
                        call descset( desc1, sz, NRHS, MB, NB, RSRC, CSRC, -1, 1 )
                    else
                        LOCR = numroc(   sz, MB, myrow(level+1), RSRC, nprow(level+1) )
                        LOCC = numroc( NRHS, NB, mycol(level+1), CSRC, npcol(level+1) )
                        allocate(gbBT(kid)%ary1(LOCR,LOCC))
                        gbBT(kid)%ary1 = ZERO
                        call descset( desc1, sz, NRHS, MB, NB, RSRC, CSRC, ctxt(level+1), max(1,LOCR) )
                    end if
                end if

                ! desc2
                LOCR = numroc(   sz, MB, myrow(level), RSRC, nprow(level) )
                LOCC = numroc( NRHS, NB, mycol(level), CSRC, npcol(level) )                
                allocate(gbBT(kid)%ary2(LOCR,LOCC))
                gbBT(kid)%ary2 = ZERO
                call descset( desc2, sz, NRHS, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )

                ! backward extend-add for the RHS
                if (level <= slvl) then
                    call pextadd_compr( -1, NRHS, gbB(level)%Q1, descB(:,1), gbB(level)%Q2, descB(:,2), gbBT(Ntmp)%ary1, descBT, gbBT(kid)%ary2, desc2, Ntmp, kid, gbNB )
                else
                    call pextadd_RHS( -1, NRHS, gbB(level)%ary, descB(:,1), gbBT(Ntmp)%ary1, descBT, gbBT(kid)%ary2, desc2, Ntmp, kid, gbNB )
                end if

                ! redistribution
                call pcgemr2d( sz, NRHS, gbBT(kid)%ary2, 1, 1, desc2, gbBT(kid)%ary1, 1, 1, desc1, ctxt(level) )
                if (allocated(gbBT(kid)%ary2)) then
                    deallocate(gbBT(kid)%ary2)
                end if

            end do

            ! deallocate gbBT when not needed
            if (allocated(gbBT(Ntmp)%ary1)) then
                deallocate(gbBT(Ntmp)%ary1)
            end if      

            if (stat /= 0) then
                if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                    flops = flops + dble(sz) * dble(NRHS) * 2.0
                end if
            end if       

        end do
      

        return
    end subroutine gb_sol


    subroutine pextadd_RHS( FOB, NRHS, B, descB, BT, descBT, X, descX, Ntmp, kid, nbr )

        !! FOB stands for "Forward Or Backward"
        integer                     :: FOB, NRHS, Ntmp, kid
        complex(kind=4),allocatable :: B(:,:), BT(:,:), X(:,:)
        integer                     :: descB(9), descBT(9), descX(9)
        type(NB_struct),target      :: nbr(*)

        integer                     :: szUL, length, counter, ii
        integer,allocatable         :: ind(:), map(:)
        type(NB_struct),pointer     :: tailN, tailK

        szUL = descB(3)
        ! length of the ind and map
        length = 0
        tailK => nbr(kid)%next
        do while (associated(tailK))
            length = length + 1
            tailK => tailK%next
        end do

        ! B
        if (FOB > 0) then
            call pcgeadd( 'N', szUL, NRHS, ONE, X, 1, 1, descX, ONE, B, 1, 1, descB )
        else
            call pcgeadd( 'N', szUL, NRHS, ONE, B, 1, 1, descB, ONE, X, 1, 1, descX )
        end if

        if (length > 0) then

            ! ind: index of each block in U
            allocate(ind(length+1))
            ind(1) = szUL + 1
            tailK => nbr(kid)%next
            do ii = 1,length
                ind(ii+1) = ind(ii) + tailK%Otail - tailK%Ohead + 1
                tailK => tailK%next
            end do

            ! mapping to the current frontal matrix
            allocate(map(length))
            tailK => nbr(kid)%next
            do ii = 1,length
                counter = 1
                tailN => nbr(Ntmp)
                do while (associated(tailN))
                    if (tailK%Ohead >= tailN%Ohead .and. tailK%Otail <= tailN%Otail) then
                        map(ii) = tailK%Ohead - tailN%Ohead + counter
                        exit
                    end if
                    counter = counter + tailN%Otail - tailN%Ohead + 1
                    tailN => tailN%next
                end do
                tailK => tailK%next
            end do

            ! BT
            do ii = 1,length
                if (FOB == 1) then
                    call pcgeadd( 'N', ind(ii+1)-ind(ii), NRHS, ONE, X, ind(ii), 1, descX, ONE, BT, map(ii), 1, descBT )
                else
                    call pcgeadd( 'N', ind(ii+1)-ind(ii), NRHS, ONE, BT, map(ii), 1, descBT, ONE, X, ind(ii), 1, descX )
                end if
            end do
            deallocate(ind,map)

        end if

        return
    end subroutine pextadd_RHS 


    subroutine pextadd_compr( FOB, NRHS, B1, descB1, B2, descB2, BT, descBT, X, descX, Ntmp, kid, nbr )

        !! FOB stands for "Forward Or Backward"
        integer                     :: FOB, NRHS, Ntmp, kid
        complex(kind=4),allocatable :: B1(:,:), B2(:,:), BT(:,:), X(:,:)
        integer                     :: descB1(9), descB2(9), descBT(9), descX(9)
        type(NB_struct),target      :: nbr(*)

        integer                     :: sz1, sz2, length, counter, ii
        integer,allocatable         :: ind(:), map(:)
        type(NB_struct),pointer     :: tailN, tailK

        sz1 = descB1(3)
        sz2 = descB2(3)
        ! length of the ind and map
        length = 0
        tailK => nbr(kid)%next
        do while (associated(tailK))
            length = length + 1
            tailK => tailK%next
        end do

        ! B
        if (FOB > 0) then
            call pcgeadd( 'N', sz1, NRHS, ONE, X,     1, 1, descX, ONE, B1, 1, 1, descB1 )
            call pcgeadd( 'N', sz2, NRHS, ONE, X, sz1+1, 1, descX, ONE, B2, 1, 1, descB2 )
        else
            call pcgeadd( 'N', sz1, NRHS, ONE, B1, 1, 1, descB1, ONE, X,     1, 1, descX )
            call pcgeadd( 'N', sz2, NRHS, ONE, B2, 1, 1, descB2, ONE, X, sz1+1, 1, descX )
        end if

        if (length > 0) then

            ! ind: index of each block in U
            allocate(ind(length+1))
            ind(1) = sz1 + sz2 + 1
            tailK => nbr(kid)%next
            do ii = 1,length
                ind(ii+1) = ind(ii) + tailK%Otail - tailK%Ohead + 1
                tailK => tailK%next
            end do

            ! mapping to the current frontal matrix
            allocate(map(length))
            tailK => nbr(kid)%next
            do ii = 1,length
                counter = 1
                tailN => nbr(Ntmp)
                do while (associated(tailN))
                    if (tailK%Ohead >= tailN%Ohead .and. tailK%Otail <= tailN%Otail) then
                        map(ii) = tailK%Ohead - tailN%Ohead + counter
                        exit
                    end if
                    counter = counter + tailN%Otail - tailN%Ohead + 1
                    tailN => tailN%next
                end do
                tailK => tailK%next
            end do

            ! BT
            do ii = 1,length
                if (FOB == 1) then
                    call pcgeadd( 'N', ind(ii+1)-ind(ii), NRHS, ONE, X, ind(ii), 1, descX, ONE, BT, map(ii), 1, descBT )
                else
                    call pcgeadd( 'N', ind(ii+1)-ind(ii), NRHS, ONE, BT, map(ii), 1, descBT, ONE, X, ind(ii), 1, descX )
                end if
            end do
            deallocate(ind,map)

        end if

        return
    end subroutine pextadd_compr



    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  3. local solution: Ux = y  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine loc_solU( slvl, Lpar, flag, lenlocTT, locTT, locTlabel, locNB, locUL, locUR, locB, gbBT, NRHS, stat, flops )

        integer                         :: slvl, Lpar, flag, lenlocTT, NRHS
        type(TT_struct)                 :: locTT(lenlocTT)
        integer                         :: locTlabel(lenlocTT+1)
        type(NB_struct),target          :: locNB(lenlocTT)
        type(UL_struct)                 :: locUL(lenlocTT)
        type(UR_struct)                 :: locUR(lenlocTT)
        complex(kind=4),allocatable     :: locB(:,:)
        type(redist_struct)             :: gbBT(*)
        integer                         :: stat
        real(kind=8)                    :: flops

        integer                     :: level, Ntmp, Nhead, Ntail, Ptmp, Phead, Ptail, szUL, szLR, rkUR, counter, info
        type(NB_struct),pointer     :: tail
        complex(kind=4),allocatable :: tmp(:,:)

        do Ntmp = lenlocTT,1,-1

            level = locTT(Ntmp)%LL
            Nhead = locTlabel(Ntmp)
            Ntail = locTlabel(Ntmp+1)-1
            szUL  = Ntail - Nhead + 1

            if (locNB(Ntmp)%Ohead /= 0) then

                if (level <= slvl-Lpar+1) then
                    rkUR = locUR(Ntmp)%rk
                    allocate(tmp(rkUR,NRHS))
                    tmp = ZERO
                end if

                counter = 1
                tail => locNB(Ntmp)
                do while (associated(tail))
                    Ptmp = tail%Node
                    szLR = tail%Otail - tail%Ohead + 1
                    if (Ptmp /= 0) then
                        Phead = tail%Ohead
                        Ptail = tail%Otail
                        if (level <= slvl-Lpar+1) then
                            call cgemm( 'N', 'N', rkUR, NRHS, szLR,  ONE,   locUR(Ntmp)%R(:,counter:counter+szLR-1), rkUR, locB(Phead:Ptail,:), szLR, ONE,                 tmp, rkUR )
                        else
                            call cgemm( 'N', 'N', szUL, NRHS, szLR, -ONE, locUR(Ntmp)%ary(:,counter:counter+szLR-1), szUL, locB(Phead:Ptail,:), szLR, ONE, locB(Nhead:Ntail,:), szUL )
                        end if
                    else
                        Phead = tail%Ohead - locTlabel(lenlocTT+1) + 1
                        Ptail = tail%Otail - locTlabel(lenlocTT+1) + 1
                        if (level <= slvl-Lpar+1) then
                            call cgemm( 'N', 'N', rkUR, NRHS, szLR,  ONE,   locUR(Ntmp)%R(:,counter:counter+szLR-1), rkUR, gbBT(flag)%ary1(Phead:Ptail,:), szLR, ONE,                 tmp, rkUR )
                        else
                            call cgemm( 'N', 'N', szUL, NRHS, szLR, -ONE, locUR(Ntmp)%ary(:,counter:counter+szLR-1), szUL, gbBT(flag)%ary1(Phead:Ptail,:), szLR, ONE, locB(Nhead:Ntail,:), szUL )
                        end if
                    end if
                    counter = counter + szLR
                    tail => tail%next
                end do
                if (stat /= 0) then
                    if (level <= slvl-Lpar+1) then
                        flops = flops + dble(rkUR) * dble(NRHS) * 2.0*dble(szLR)
                    else 
                        flops = flops + dble(szUL) * dble(NRHS) * 2.0*dble(szLR)
                    end if
                end if

                if (level <= slvl-Lpar+1) then
                    call cgemm( 'N', 'N', szUL, NRHS, rkUR, -ONE, locUR(Ntmp)%Q1, szUL, tmp, rkUR, ONE, locB(Nhead:Ntail,:), szUL )
                    deallocate(tmp)
                    if (stat /= 0) then
                        flops = flops + dble(szUL) * dble(NRHS) * 2.0*dble(rkUR)
                    end if
                end if

            end if

            if (level <= slvl-Lpar+1) then
                call loctrs_compr( 1, 0, NRHS, locUL(Ntmp), locIPIV(Ntmp), locB(Nhead:Ntail,1:NRHS), szUL )
            else
                call ctrtrs( 'U', 'N', 'N', szUL, NRHS, locUL(Ntmp)%ary11, szUL, locB(Nhead:Ntail,1:NRHS), szUL, info )
            end if
            if (stat /= 0) then
                flops = flops + dble(szUL)**2.0 * dble(NRHS)
            end if

        end do

        return
    end subroutine loc_solU


end module sol_kernel
