module solution2D

use sol_kernel
implicit none
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  solution module: loc_sol + gb_sol  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


contains


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  2D solver part  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine mfsolve(X,NRHS,slvl,Lpar,flag,Nz,Nx,locHZ,locTZ,locHX,locTX,label,ctxt,nprow,npcol,myrow,mycol,mynodes, &
                       lenlocTT,lengbTT,locTT,gbTT,locTlabel,gbTlabel,locNB,gbNB,locUL,locUR,locLL,locIPIV,gbUL,gbUR,gbLL,gbIPIV)

        complex(kind=4)                 :: X(locTZ-locHZ+1,locTX-locHX+1,NRHS)
        integer                         :: NRHS, slvl, Lpar, flag, Nz, Nx, locHZ, locTZ, locHX, locTX, lenlocTT, lengbTT
        integer                         :: label(locTZ-locHZ+1,locTX-locHX+1)
        integer                         :: ctxt(Lpar), nprow(Lpar), npcol(Lpar), myrow(Lpar), mycol(Lpar), mynodes(Lpar)
        integer                         :: locTlabel(lenlocTT+1), gbTlabel(lengbTT+1)
        type(TT_struct)                 :: locTT(lenlocTT), gbTT(lengbTT)
        type(NB_struct)                 :: locNB(lenlocTT), gbNB(lengbTT)
        type(UL_struct)                 :: locUL(lenlocTT), gbUL(lengbTT)
        type(UR_struct)                 :: locUR(lenlocTT), locLL(lenlocTT), gbUR(lengbTT), gbLL(lengbTT)
        type(ipiv_struct)               :: locIPIV(lenlocTT), gbIPIV(lengbTT)

        complex(kind=4),allocatable     :: locB(:,:)
        type(redist_struct),allocatable :: gbBT(:)


        !! 1. generate locB and gbBT
        call locRHS_generate( X, NRHS, flag, locHZ, locTZ, locHX, locTX, label, lenlocTT, lengbTT, locTlabel, locB, gbBT )


        !! 2. multifrontal solver kernel
        call mfsolve_kernel(locUL,locUR,locLL,locIPIV,gbUL,gbUR,gbLL,gbIPIV,slvl,Lpar,flag,ctxt,nprow,npcol,myrow,mycol,mynodes, & 
                            lenlocTT,lengbTT,locTT,gbTT,locTlabel,gbTlabel,locNB,gbNB,NRHS,locB,gbBT)


        !! 3. obtain the final solution
        call locX_generate( X, NRHS, flag, locHZ, locTZ, locHX, locTX, label, lenlocTT, lengbTT, locTlabel, locB, gbBT )
        deallocate(locB,gbBT(flag)%ary1)
        deallocate(gbBT)


        return
    end subroutine mfsolve


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  generate locB and gbBT  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine locRHS_generate( X, NRHS, flag, locHZ, locTZ, locHX, locTX, label, lenlocTT, lengbTT, locTlabel, locB, gbBT )

        integer                         :: NRHS, flag, locHZ, locTZ, locHX, locTX, lenlocTT, lengbTT
        integer                         :: label(locTZ-locHZ+1,locTX-locHX+1), locTlabel(lenlocTT+1)
        complex(kind=4)                 :: X(locTZ-locHZ+1,locTX-locHX+1,NRHS)
        complex(kind=4),allocatable     :: locB(:,:)
        type(redist_struct),allocatable :: gbBT(:)

        integer      :: rhs, zz, xx

        allocate( gbBT(lengbTT) )
        allocate( locB(locTlabel(lenlocTT+1)-1,NRHS) )
        allocate( gbBT(flag)%ary1( (locTZ-locHZ+1)*(locTX-locHX+1)-locTlabel(lenlocTT+1)+1, NRHS ) )

        do rhs = 1,NRHS
            do xx = 1,locTX-locHX+1
                do zz = 1,locTZ-locHZ+1
                    if (label(zz,xx) >= locTlabel(lenlocTT+1)) then
                        gbBT(flag)%ary1( label(zz,xx)-locTlabel(lenlocTT+1)+1, rhs ) = X(zz,xx,rhs)
                    else
                        locB(label(zz,xx),rhs) = X(zz,xx,rhs)
                    end if
                end do
            end do
        end do

        return
    end subroutine locRHS_generate


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  obtain the final solution in X  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine locX_generate( X, NRHS, flag, locHZ, locTZ, locHX, locTX, label, lenlocTT, lengbTT, locTlabel, locB, gbBT )

        integer                         :: NRHS, flag, locHZ, locTZ, locHX, locTX, lenlocTT, lengbTT
        integer                         :: label(locTZ-locHZ+1,locTX-locHX+1), locTlabel(lenlocTT+1)
        complex(kind=4)                 :: X(locTZ-locHZ+1,locTX-locHX+1,NRHS)
        complex(kind=4),allocatable     :: locB(:,:)
        type(redist_struct)             :: gbBT(lengbTT)

        integer :: rhs, zz, xx

        do rhs = 1,NRHS
            do xx = 1,locTX-locHX+1
                do zz = 1,locTZ-locHZ+1
                    if (label(zz,xx) >= locTlabel(lenlocTT+1)) then
                        X(zz,xx,rhs) = gbBT(flag)%ary1( label(zz,xx)-locTlabel(lenlocTT+1)+1, rhs )
                    else
                        X(zz,xx,rhs) = locB(label(zz,xx),rhs)
                    end if
                end do
            end do
        end do

        return
    end subroutine locX_generate


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  write X to a file  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine cwriteX( ROI, filename, X, NRHS, tos, Nz, Nx, locHZ, locTZ, locHX, locTX )

        integer                     :: NRHS, ROI    ! Real Or Imag ?
        character(len=200)          :: filename
        complex(kind=4)             :: X(locTZ-locHZ+1,locTX-locHX+1,NRHS)
        integer                     :: tos, Nz, Nx, locHZ, locTZ, locHX, locTX
        integer                     :: counter, rhs, zz, xx, headZ, headX

        headZ = locHZ
        if (headZ > 1) then
            headZ = headZ + tos
        end if
        headX = locHX
        if (headX > 1) then
            headX = headX + tos
        end if

        open( unit = 10, file = filename, form = "unformatted", status = "old", access = "direct", recl = 1 )

        if (ROI == 1) then    ! real part
            do rhs = 1,NRHS
                do xx = headX,locTX
                    do zz = headZ,locTZ
                        counter = (rhs-1)*Nz*Nx + (xx-1)*Nz + zz
                        write(10,rec=counter) real(X(zz-locHZ+1,xx-locHX+1,rhs))
                    end do
                end do
            end do
        elseif (ROI == -1) then   ! imag part
            do rhs = 1,NRHS
                do xx = headX,locTX
                    do zz = headZ,locTZ
                        counter = (rhs-1)*Nz*Nx + (xx-1)*Nz + zz
                        write(10,rec=counter) aimag(X(zz-locHZ+1,xx-locHX+1,rhs))
                    end do
                end do
            end do
        elseif (ROI == 0) then    ! both parts
            do rhs = 1,NRHS
                do xx = headX,locTX
                    do zz = headZ,locTZ
                        counter = ((rhs-1)*Nz*Nx+(xx-1)*Nz+(zz-1)) * 2 + 1
                        write(10,rec=counter)    real(X(zz-locHZ+1,xx-locHX+1,rhs))
                        write(10,rec=counter+1) aimag(X(zz-locHZ+1,xx-locHX+1,rhs))
                    end do
                end do
            end do
        end if

        close(10)
        
        return
    end subroutine cwriteX


    subroutine swriteX( filename, X, NRHS, tos, Nz, Nx, locHZ, locTZ, locHX, locTX )

        character(len=200)          :: filename
        real(kind=4)                :: X(locTZ-locHZ+1,locTX-locHX+1,NRHS)
        integer                     :: NRHS, tos, Nz, Nx, locHZ, locTZ, locHX, locTX
        integer                     :: counter, rhs, zz, xx, headZ, headX

        headZ = locHZ
        if (headZ > 1) then
            headZ = headZ + tos
        end if
        headX = locHX
        if (headX > 1) then
            headX = headX + tos
        end if

        open( unit = 10, file = filename, form = "unformatted", status = "old", access = "direct", recl = 1 )
        do rhs = 1,NRHS
            do xx = headX,locTX
                do zz = headZ,locTZ
                    counter = (rhs-1)*Nz*Nx + (xx-1)*Nz + zz
                    write(10,rec=counter) X(zz-locHZ+1,xx-locHX+1,rhs)
                end do
            end do
        end do
        close(10)
        
        return
    end subroutine swriteX


end module solution2D
